(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var SandstormDb, Packages, DevPackages, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, AppIndex, KeybaseProfiles, isDemoUser, isSignedUp, isSignedUpOrDemo, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, roleAssignmentPattern, identityIds, i, appName, addStaticAsset, userPictureUrl, fetchPicture;

(function () {

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/db.js                                                                               //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
// This file defines the database schema.                                                                    // 17
                                                                                                             // 18
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every               // 19
// query it makes, so you can see if it's doing queries too often, etc.                                      // 20
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                                      // 21
  var oldFind = Mongo.Collection.prototype.find;                                                             // 22
  Mongo.Collection.prototype.find = function () {                                                            // 23
    console.log(this._prefix, arguments);                                                                    // 24
    return oldFind.apply(this, arguments);                                                                   // 25
  }                                                                                                          // 26
}                                                                                                            // 27
                                                                                                             // 28
// Users = new Mongo.Collection("users");                                                                    // 29
// The users collection is special and can be accessed through `Meteor.users`.                               // 30
// See https://docs.meteor.com/#/full/meteor_users. Entries in the users collection correspond to            // 31
// accounts. Each represents a capability store belonging to a human.                                        // 32
//                                                                                                           // 33
// Note that accounts are not the same thing as identities. An account may have multiple identities          // 34
// attached to it. Identities provide a way to authenticate the user and are also used to present            // 35
// globally unique and stable user indentifiers to grains via `Grain.UserInfo.identityId`.                   // 36
//                                                                                                           // 37
// Each entry in this collection contains:                                                                   // 38
//   _id: Random string. What we're talking about when we say "User ID" or "Account ID".                     // 39
//   createdAt: Date when this entry was added to the collection.                                            // 40
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.                      // 41
//   profile: Obsolete now that we allow more than one identity per account.                                 // 42
//   identities: Array of identity objects, each of which may include the following fields.                  // 43
//       id: The globally-stable SHA-256 ID of this identity. This field must be present.                    // 44
//       service: Object specifying how to authenticate as this identity. The object contains a              // 45
//                single key, representing the name of the service, corresponding to the `type`              // 46
//                field in `Accounts.validateLoginAttempt()`. The value associated with the key              // 47
//                is an object containing provider-specific data, e.g. for the `emailToken`                  // 48
//                service it contains the active tokens.                                                     // 49
//       unverifiedEmail: Email address specified by the user.                                               // 50
//       verifiedEmail: Only provided by some services. Cannot be directly edited by the user.               // 51
//       main: True if this is the user's main identity.                                                     // 52
//       noLogin: True if the user does not trust this identity for account authentication.                  // 53
//       profile: Object containing the data that will be shared with users grains that come into            // 54
//                contact with this identity. May include the following fields.                              // 55
//           name: String containing the display name of the identity. Default: first part of email.         // 56
//           handle: String containing the identity's preferred handle. Default: first part of email.        // 57
//           picture: _id into the StaticAssets table for the identity's picture. Default: identicon.        // 58
//           pronoun: One of "male", "female", "neutral", or "robot". Default: neutral.                      // 59
//   services: Object containing login and identity data used by Meteor authentication services.             // 60
//   mergedUsers: Array of User _id strings, representing the accounts that have been merged into this       // 61
//                one. Those accounts remain in the Users collection, stripped of their `identities`         // 62
//                and `services` fields.                                                                     // 63
//   isAdmin: Boolean indicating whether this user is allowed to access the Sandstorm admin panel.           // 64
//   signupKey: If this is an invited user, then this field contains their signup key.                       // 65
//   signupNote: If the user was invited through a link, then this field contains the note that the          // 66
//               inviter admin attached to the key.                                                          // 67
//   signupEmail: If the user was invited by email, then this field contains the email address that          // 68
//                the invite was sent to.                                                                    // 69
//   plan: _id of an entry in the Plans table which determines the user's qutoa.                             // 70
//   storageUsage: Number of bytes this user is currently storing.                                           // 71
//   expires: Date when this user's account should be deleted. Only present for demo users.                  // 72
//   isAppDemoUser: True if this is a demo user who arrived via an /appdemo/ link.                           // 73
//   appDemoId: If this is an appdemo user (see above), the app ID they started out demoing.                 // 74
//   payments: Object defined by payments module, if loaded.                                                 // 75
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                       // 76
                                                                                                             // 77
Packages = new Mongo.Collection("packages");                                                                 // 78
// Packages which are installed or downloading.                                                              // 79
//                                                                                                           // 80
// Each contains:                                                                                            // 81
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                          // 82
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"         // 83
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),                      // 84
//       >1 = download byte count.                                                                           // 85
//   error:  If status is "failed", error message string.                                                    // 86
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.                // 87
//   appId:  If status is "ready", the application ID string.  Packages representing different               // 88
//       versions of the same app have the same appId.  The spk tool defines the app ID format               // 89
//       and can cryptographically verify that a package belongs to a particular app ID.                     // 90
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package              // 91
//       collector should at some point check whether there are any other references and, if not,            // 92
//       delete the package.                                                                                 // 93
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.               // 94
//   isAutoUpdated: This package was downloaded as part of an auto-update. We shouldn't clean it up          // 95
//     even if it has no users.                                                                              // 96
//   authorPgpKeyFingerprint: Verified PGP key fingerprint (SHA-1, hex, all-caps) of the app                 // 97
//     packager.                                                                                             // 98
                                                                                                             // 99
DevPackages = new Mongo.Collection("devpackages");                                                           // 100
// List of packages currently made available via the dev tools running on the local machine.                 // 101
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool           // 102
// on the local machine to publish an under-development app to this server. That should only ever            // 103
// happen on developers' desktop machines.                                                                   // 104
//                                                                                                           // 105
// While a dev package is published, it automatically appears as installed by every user of the              // 106
// server, and it overrides all packages with the same application ID. If any instances of those             // 107
// packages are currently open, they are killed and reset on publish.                                        // 108
//                                                                                                           // 109
// When the dev tool disconnects, the package is automatically unpublished, and any open instances           // 110
// are again killed and refreshed.                                                                           // 111
//                                                                                                           // 112
// Each contains:                                                                                            // 113
//   _id:  The package ID string (as with Packages._id).                                                     // 114
//   appId: The app ID this package is intended to override (as with Packages.appId).                        // 115
//   timestamp:  Time when the package was last updated. If this changes while the package is                // 116
//     published, all running instances are reset. This is used e.g. to reset the app each time              // 117
//     changes are made to the source code.                                                                  // 118
//   manifest:  The app's manifest, as with Packages.manifest.                                               // 119
                                                                                                             // 120
UserActions = new Mongo.Collection("userActions");                                                           // 121
// List of actions that each user has installed which create new grains.  Each app may install               // 122
// some number of actions (usually, one).                                                                    // 123
//                                                                                                           // 124
// Each contains:                                                                                            // 125
//   _id:  random                                                                                            // 126
//   userId:  User who has installed this action.                                                            // 127
//   packageId:  Package used to run this action.                                                            // 128
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 129
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                       // 130
//       that clients can access it without subscribing to the Packages collection.                          // 131
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 132
//       searchability.                                                                                      // 133
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"                     // 134
//   title: JSON-encoded LocalizedText title for this action, e.g.                                           // 135
//       `{defaultText: "New Spreadsheet"}`.                                                                 // 136
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.              // 137
//   command:  Manifest.Command to run this action (see package.capnp).                                      // 138
                                                                                                             // 139
Grains = new Mongo.Collection("grains");                                                                     // 140
// Grains belonging to users.                                                                                // 141
//                                                                                                           // 142
// Each contains:                                                                                            // 143
//   _id:  random                                                                                            // 144
//   packageId:  _id of the package of which this grain is an instance.                                      // 145
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 146
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 147
//       searchability.                                                                                      // 148
//   userId: The _id of the user who owns this grain.                                                        // 149
//   identityId: Identity of user who owns this grain.                                                       // 150
//   title:  Human-readable string title, as chosen by the user.                                             // 151
//   lastUsed:  Date when the grain was last used by a user.                                                 // 152
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                          // 153
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent          // 154
//                   time a session to this grain was opened.                                                // 155
//                                                                                                           // 156
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and           // 157
// web publishing functionality without powerbox support; they will be replaced once the powerbox            // 158
// is implemented.                                                                                           // 159
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and           // 160
//       web publishing. This field is initialized when first requested by the app.                          // 161
                                                                                                             // 162
RoleAssignments = new Mongo.Collection("roleAssignments");                                                   // 163
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges          // 164
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens                   // 165
// collection.                                                                                               // 166
                                                                                                             // 167
Contacts = new Mongo.Collection("contacts");                                                                 // 168
// Edges in the social graph.                                                                                // 169
//                                                                                                           // 170
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob               // 171
// will show up in her user-picker UI for actions like share-by-identity.                                    // 172
//                                                                                                           // 173
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of              // 174
// Bob's.                                                                                                    // 175
//                                                                                                           // 176
// Each contains:                                                                                            // 177
//   _id: random                                                                                             // 178
//   ownerId: The `_id` of the user who owns this contact.                                                   // 179
//   identityId:  The identity of the contacted user.                                                        // 180
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies              // 181
//            the contact to the owner.                                                                      // 182
//   created: Date when this contact was created.                                                            // 183
                                                                                                             // 184
Sessions = new Mongo.Collection("sessions");                                                                 // 185
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.          // 186
//                                                                                                           // 187
// Each contains:                                                                                            // 188
//   _id:  random                                                                                            // 189
//   grainId:  _id of the grain to which this session is connected.                                          // 190
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the           // 191
//       '*' in WILDCARD_HOST.                                                                               // 192
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some              // 193
//       period.                                                                                             // 194
//   userId:  User ID of the user who owns this session.                                                     // 195
//   identityId:  Identity ID of the user who owns this session.                                             // 196
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens            // 197
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),               // 198
//       anonymous users can get access without an API token and so neither userId nor hashedToken           // 199
//       are present.                                                                                        // 200
//   powerboxView: If present, this is a view that should be presented as part of a powerbox                 // 201
//       interaction.                                                                                        // 202
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                         // 203
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated           // 204
//       until newSession is called on the UiView.                                                           // 205
//   permissions: The permissions for the current identity on this UiView. This isn't populated              // 206
//       until newSession is called on the UiView.                                                           // 207
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first         // 208
//       request                                                                                             // 209
                                                                                                             // 210
SignupKeys = new Mongo.Collection("signupKeys");                                                             // 211
// Invite keys which may be used by users to get access to Sandstorm.                                        // 212
//                                                                                                           // 213
// Each contains:                                                                                            // 214
//   _id:  random                                                                                            // 215
//   used:  Boolean indicating whether this key has already been consumed.                                   // 216
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.                // 217
//   email: If this key was sent as an email invite, the email address to which it was sent.                 // 218
                                                                                                             // 219
ActivityStats = new Mongo.Collection("activityStats");                                                       // 220
// Contains usage statistics taken on a regular interval. Each entry is a data point.                        // 221
//                                                                                                           // 222
// Each contains:                                                                                            // 223
//   timestamp: Date when measurements were taken.                                                           // 224
//   daily: Contains stats counts pertaining to the last day before the sample time.                         // 225
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.                 // 226
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.                 // 227
//                                                                                                           // 228
// Each of daily, weekly, and monthly contains:                                                              // 229
//   activeUsers: The number of unique users who have used a grain on the server in the time                 // 230
//       interval. Only counts logged-in users.                                                              // 231
//   demoUsers: Demo users.                                                                                  // 232
//   appDemoUsers: Users that came in through "app demo".                                                    // 233
//   activeGrains: The number of unique grains that have been used in the time interval.                     // 234
//   apps: An object indexed by app ID recording, for each app:                                              // 235
//       owners: Number of unique owners of this app (counting only grains that still exist).                // 236
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting          // 237
//         only grains that still exist).                                                                    // 238
//       grains: Number of active grains of this app (that still exist).                                     // 239
//       deleted: Number of non-demo grains of this app that were deleted.                                   // 240
//       demoed: Number of demo grains created and expired.                                                  // 241
//       appDemoUsers: Number of app demos initiated with this app.                                          // 242
                                                                                                             // 243
DeleteStats = new Mongo.Collection("deleteStats");                                                           // 244
// Contains records of objects that were deleted, for stat-keeping purposes.                                 // 245
//                                                                                                           // 246
// Each contains:                                                                                            // 247
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                                   // 248
//   lastActive: Date of the user's or grain's last activity.                                                // 249
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they           // 250
//     arrived to demo. For others, undefined.                                                               // 251
                                                                                                             // 252
FileTokens = new Mongo.Collection("fileTokens");                                                             // 253
// Tokens corresponding to backup files that are currently stored on the server. A user receives             // 254
// a token when they create a backup file (either by uploading it, or by backing up one of their             // 255
// grains) and may use the token to read the file (either to download it, or to restore a new                // 256
// grain from it).                                                                                           // 257
//                                                                                                           // 258
// Each contains:                                                                                            // 259
//   _id:       The unguessable token string.                                                                // 260
//   name:      Suggested filename.                                                                          // 261
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.              // 262
                                                                                                             // 263
ApiTokens = new Mongo.Collection("apiTokens");                                                               // 264
// Access tokens for APIs exported by apps.                                                                  // 265
//                                                                                                           // 266
// Originally API tokens were only used by external users through the HTTP API endpoint. However,            // 267
// now they are also used to implement SturdyRefs, not just held by external users, but also when            // 268
// an app holds a SturdyRef to another app within the same server. See the various `save()`,                 // 269
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and               // 270
// `MainView`) -- the fields of type `Data` are API tokens.                                                  // 271
//                                                                                                           // 272
// Each contains:                                                                                            // 273
//   _id:       A SHA-256 hash of the token, base64-encoded.                                                 // 274
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)            // 275
//   identityId: For UiView capabilities, this is the identity for which the view is attenuated.             // 276
//              That is, the UiView's newSession() method will intersect the requested permissions           // 277
//              with this identity's permissions before forwarding on to the underlying app. If              // 278
//              `identityId` is not present, then no identity attenuation is applied, i.e. this is           // 279
//              a raw UiView as implemented by the app. (The `roleAssignment` field, below, may              // 280
//              still apply. For non-UiView capabilities, `identityId` is never present. Note that           // 281
//              this is NOT the identity against which the `requiredPermissions` parameter of                // 282
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerIdentity`.         // 283
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded               // 284
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These          // 285
//              permissions will be intersected with those held by `identityId` when the view is             // 286
//              opened.                                                                                      // 287
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as          // 288
//              anonymous rather than as directly associated with `identityId`. This has no effect           // 289
//              on the permissions granted.                                                                  // 290
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by            // 291
//              the app or its supervisor (whereas without this it strictly represents UiView).              // 292
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).            // 293
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is                   // 294
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.                     // 295
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,          // 296
//              not a particular grain. (`grainId` and `identityId` are not set.) This is an object          // 297
//              containing exactly one of the following fields:                                              // 298
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                          // 299
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the                 // 300
//                           `Notifications` collection.                                                     // 301
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this            // 302
//                  will be moved out of the frontend and into the backend, but we'll migrate the            // 303
//                  database when that happens. This field contains the boolean true to signify that         // 304
//                  it has been set.                                                                         // 305
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                           // 306
//   parentToken: If present, then this token represents exactly the capability represented by               // 307
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated          // 308
//              by `roleAssignment` (if present). To facilitate permissions computations, if the             // 309
//              capability is a UiView, then `grainId` is set to the backing grain and `identityId`          // 310
//              is set to the identity that shared the view. Neither `objectId` nor `frontendRef`            // 311
//              is present when `parentToken` is present.                                                    // 312
//   petname:   Human-readable label for this access token, useful for identifying tokens for                // 313
//              revocation. This should be displayed when visualizing incoming capabilities to               // 314
//              the grain identified by `grainId`.                                                           // 315
//   created:   Date when this token was created.                                                            // 316
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may          // 317
//              become un-revoked in the future.                                                             // 318
//   expires:   Optional expiration Date. If undefined, the token does not expire.                           // 319
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)                // 320
//              as passed to the `save()` call that created this token. If not present, treat                // 321
//              as `webkey` (the default for `ApiTokenOwner`).                                               // 322
//   expiresIfUnused:                                                                                        // 323
//              Optional Date after which the token, if it has not been used yet, expires.                   // 324
//              This field should be cleared on a token's first use.                                         // 325
//   requirements: List of conditions which must hold for this token to be considered valid.                 // 326
//              Semantically, this list specifies the powers which were *used* to originally                 // 327
//              create the token. If any condition in the list becomes untrue, then the token must           // 328
//              be considered revoked, and all live refs and sturdy refs obtained transitively               // 329
//              through it must also become revoked. Each item is the JSON serialization of the              // 330
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                               // 331
//                                                                                                           // 332
// It is important to note that a token's owner and provider are independent from each other. To             // 333
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:          // 334
//                                                                                                           // 335
// struct ApiToken {                                                                                         // 336
//   owner :ApiTokenOwner;                                                                                   // 337
//   provider :union {                                                                                       // 338
//     grain :group {                                                                                        // 339
//       grainId :Text;                                                                                      // 340
//       union {                                                                                             // 341
//         uiView :group {                                                                                   // 342
//           identityId :Text;                                                                               // 343
//           roleAssignment :RoleAssignment;                                                                 // 344
//           forSharing :Bool;                                                                               // 345
//         }                                                                                                 // 346
//         objectId :SupervisorObjectId;                                                                     // 347
//       }                                                                                                   // 348
//     }                                                                                                     // 349
//     frontendRef :union {                                                                                  // 350
//        notificationHandle :Text;                                                                          // 351
//        ipNetwork :Bool;                                                                                   // 352
//        ipInterface :Bool;                                                                                 // 353
//     }                                                                                                     // 354
//     child :group {                                                                                        // 355
//       parentToken :Text;                                                                                  // 356
//       union {                                                                                             // 357
//         uiView :group {                                                                                   // 358
//           grainId :Text;                                                                                  // 359
//           identityId :Text;                                                                               // 360
//           roleAssignment :RoleAssignment = (allAccess = ());                                              // 361
//         }                                                                                                 // 362
//         other :Void;                                                                                      // 363
//       }                                                                                                   // 364
//     }                                                                                                     // 365
//   }                                                                                                       // 366
//   requirements: List(Supervisor.MembraneRequirement);                                                     // 367
//   ...                                                                                                     // 368
// }                                                                                                         // 369
                                                                                                             // 370
Notifications = new Mongo.Collection("notifications");                                                       // 371
// Notifications for a user.                                                                                 // 372
//                                                                                                           // 373
// Each contains:                                                                                            // 374
//   _id:          random                                                                                    // 375
//   grainId:      The grain originating this notification, if any.                                          // 376
//   userId:       The user receiving the notification.                                                      // 377
//   text:         The JSON-ified LocalizedText to display in the notification.                              // 378
//   isUnread:     Boolean indicating if this notification is unread.                                        // 379
//   timestamp:    Date when this notification was last updated                                              // 380
//   ongoing:      If present, this is an ongoing notification, and this field contains an                   // 381
//                 ApiToken referencing the `OngoingNotification` capability.                                // 382
//   admin:        If present, this is a notification intended for an admin.                                 // 383
//     action:     If present, this is a (string) link that the notification should direct the               // 384
//                 admin to.                                                                                 // 385
//     type:       The type of notification (currently only "reportStats").                                  // 386
//   appUpdates:   If present, this is an app update notification. It is an object with the appIds           // 387
//                 as keys.                                                                                  // 388
//     $appId:     The appId that has an outstanding update.                                                 // 389
//       packageId: The packageId that it will update to.                                                    // 390
//       name: The name of the app. (appTitle from package.manifest)                                         // 391
//       version: The app's version number. (appVersion from package.manifest)                               // 392
//       marketingVersion: String marketing version of this app. (appMarketingVersion from package.manifest) // 393
                                                                                                             // 394
StatsTokens = new Mongo.Collection("statsTokens");                                                           // 395
// Access tokens for the Stats collection                                                                    // 396
//                                                                                                           // 397
// These tokens are used for accessing the ActivityStats collection remotely                                 // 398
// (ie. from a dashboard webapp)                                                                             // 399
//                                                                                                           // 400
// Each contains:                                                                                            // 401
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                        // 402
                                                                                                             // 403
Misc = new Mongo.Collection("misc");                                                                         // 404
// Miscellaneous configuration and other settings                                                            // 405
//                                                                                                           // 406
// This table is currently only used for persisting BASE_URL from one session to the next,                   // 407
// but in general any miscellaneous settings should go in here                                               // 408
//                                                                                                           // 409
// Each contains:                                                                                            // 410
//   _id:       The name of the setting. eg. "BASE_URL"                                                      // 411
//   value:     The value of the setting.                                                                    // 412
                                                                                                             // 413
Settings = new Mongo.Collection("settings");                                                                 // 414
// Settings for this Sandstorm instance go here. They are configured through the adminSettings               // 415
// route. This collection differs from misc in that this collection is completely user controlled.           // 416
//                                                                                                           // 417
// Each contains:                                                                                            // 418
//   _id:       The name of the setting. eg. "MAIL_URL"                                                      // 419
//   value:     The value of the setting.                                                                    // 420
//   potentially other fields that are unique to the setting                                                 // 421
                                                                                                             // 422
Migrations = new Mongo.Collection("migrations");                                                             // 423
// This table tracks which migrations we have applied to this instance.                                      // 424
// It contains a single entry:                                                                               // 425
//   _id:       "migrations_applied"                                                                         // 426
//   value:     The number of migrations this instance has successfully completed.                           // 427
                                                                                                             // 428
StaticAssets = new Mongo.Collection("staticAssets");                                                         // 429
// Collection of static assets served up from the Sandstorm server's "static" host. We only                  // 430
// support relatively small assets: under 1MB each.                                                          // 431
//                                                                                                           // 432
// Each contains:                                                                                            // 433
//   _id:       Random ID; will be used in the URL.                                                          // 434
//   hash:      A base64-encoded SHA-256 hash of the data, used to de-dupe.                                  // 435
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                                    // 436
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                          // 437
//   content:   The asset content (byte buffer).                                                             // 438
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't          // 439
//       have transactions, this needs to bias towards over-counting; a backup GC could be used              // 440
//       to catch leaked assets, although it's probably not a big deal in practice.                          // 441
                                                                                                             // 442
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                               // 443
// Collection of tokens representing a single-use permission to upload an asset, such as a new               // 444
// profile picture.                                                                                          // 445
//                                                                                                           // 446
// Each contains:                                                                                            // 447
//   _id:       Random ID.                                                                                   // 448
//   purpose:   Contains one of the following, indicating how the asset is to be used:                       // 449
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:                // 450
//           userId: User whose picture shall be replaced.                                                   // 451
//           identityId: Which of the user's identities shall be updated.                                    // 452
//   expires:   Time when this token will go away if unused.                                                 // 453
                                                                                                             // 454
Plans = new Mongo.Collection("plans");                                                                       // 455
// Subscription plans, which determine quota.                                                                // 456
//                                                                                                           // 457
// Each contains:                                                                                            // 458
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...                      // 459
//   storage: Number of bytes this user is allowed to store.                                                 // 460
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                                // 461
//   computeLabel: Label to display to the user describing this plan's compute units.                        // 462
//   grains: Total number of grains this user can create (often `Infinity`).                                 // 463
//   price: Price per month in US cents.                                                                     // 464
                                                                                                             // 465
AppIndex = new Mongo.Collection("appIndex");                                                                 // 466
// A mirror of the data from the App Market index                                                            // 467
//                                                                                                           // 468
// Each contains:                                                                                            // 469
//   _id: the appId of the app                                                                               // 470
//  The rest of the fields are defined in src/sandstorm/app-index/app-index.capnp:AppIndexForMarket          // 471
                                                                                                             // 472
KeybaseProfiles = new Mongo.Collection("keybaseProfiles");                                                   // 473
// Cache of Keybase profile information. The profile for a user is re-fetched every time a package           // 474
// by that user is installed, as well as if the keybase profile is requested and not already                 // 475
// present for some reason.                                                                                  // 476
//                                                                                                           // 477
// Each contains:                                                                                            // 478
//   _id: PGP key fingerprint (SHA-1, hex, all-caps)                                                         // 479
//   displayName: Display name from Keybase. (NOT VERIFIED AT ALL.)                                          // 480
//   handle: Keybase handle.                                                                                 // 481
//   proofs: The "proofs_summary.all" array from the Keybase lookup. See the non-existent Keybase            // 482
//     docs for details. We also add a boolean "status" field to each proof indicating whether               // 483
//     we have directly verified the proof ourselves. Its values may be "unverified" (Keybase                // 484
//     returned this but we haven't checked it directly), "verified" (we verified the proof and it           // 485
//     is valid), "invalid" (we checked the proof and it was definitely bogus), or "checking" (the           // 486
//     server is currently actively checking this proof). Note that if a check fails due to network          // 487
//     errors, the status goes back to "unverified".                                                         // 488
//                                                                                                           // 489
//     WARNING: Currently verification is NOT IMPLEMENTED, so all proofs will be "unverified"                // 490
//       for now and we just trust Keybase.                                                                  // 491
                                                                                                             // 492
if (Meteor.isServer) {                                                                                       // 493
  Meteor.publish("credentials", function () {                                                                // 494
    // Data needed for isSignedUp() and isAdmin() to work.                                                   // 495
                                                                                                             // 496
    if (this.userId) {                                                                                       // 497
      return [                                                                                               // 498
        Meteor.users.find({_id: this.userId},                                                                // 499
            {fields: {signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,                                 // 500
                      plan: 1, hasCompletedSignup: 1}}),                                                     // 501
        Plans.find()                                                                                         // 502
      ];                                                                                                     // 503
    } else {                                                                                                 // 504
      return [];                                                                                             // 505
    }                                                                                                        // 506
  });                                                                                                        // 507
}                                                                                                            // 508
                                                                                                             // 509
isDemoUser = function() {                                                                                    // 510
  // Returns true if this is a demo user.                                                                    // 511
                                                                                                             // 512
  var user = Meteor.user();                                                                                  // 513
  if (user && user.expires) {                                                                                // 514
    return true;                                                                                             // 515
  } else {                                                                                                   // 516
    return false;                                                                                            // 517
  }                                                                                                          // 518
}                                                                                                            // 519
                                                                                                             // 520
isSignedUp = function() {                                                                                    // 521
  // Returns true if the user has presented an invite key.                                                   // 522
                                                                                                             // 523
  var user = Meteor.user();                                                                                  // 524
                                                                                                             // 525
  if (!user) return false;  // not signed in                                                                 // 526
                                                                                                             // 527
  if (user.expires) return false;  // demo user.                                                             // 528
                                                                                                             // 529
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 530
                                                                                                             // 531
  if (user.signupKey) return true;  // user is invited                                                       // 532
                                                                                                             // 533
  return false;                                                                                              // 534
}                                                                                                            // 535
                                                                                                             // 536
isSignedUpOrDemo = function () {                                                                             // 537
  var user = Meteor.user();                                                                                  // 538
                                                                                                             // 539
  if (!user) return false;  // not signed in                                                                 // 540
                                                                                                             // 541
  if (user.expires) return true;  // demo user.                                                              // 542
                                                                                                             // 543
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 544
                                                                                                             // 545
  if (user.signupKey) return true;  // user is invited                                                       // 546
                                                                                                             // 547
  return false;                                                                                              // 548
}                                                                                                            // 549
                                                                                                             // 550
isUserOverQuota = function (user) {                                                                          // 551
  // Return false if user has quota space remaining, true if it is full. When this returns true,             // 552
  // we will not allow the user to create new grains, though they may be able to open existing ones          // 553
  // which may still increase their storage usage.                                                           // 554
  //                                                                                                         // 555
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 556
                                                                                                             // 557
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 558
                                                                                                             // 559
  var plan = Plans.findOne(user.plan || "free");                                                             // 560
                                                                                                             // 561
  if (plan.grains < Infinity) {                                                                              // 562
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains}).count();                   // 563
    if (count >= plan.grains) return "outOfGrains";                                                          // 564
  }                                                                                                          // 565
                                                                                                             // 566
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";                   // 567
}                                                                                                            // 568
                                                                                                             // 569
isUserExcessivelyOverQuota = function (user) {                                                               // 570
  // Return true if user is so far over quota that we should prevent their existing grains from              // 571
  // running at all.                                                                                         // 572
  //                                                                                                         // 573
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 574
                                                                                                             // 575
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 576
                                                                                                             // 577
  var plan = Plans.findOne(user.plan || "free");                                                             // 578
                                                                                                             // 579
  if (plan.grains < Infinity) {                                                                              // 580
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains * 2}).count();               // 581
    if (count >= plan.grains * 2) return "outOfGrains";                                                      // 582
  }                                                                                                          // 583
                                                                                                             // 584
  return plan && user.storageUsage && user.storageUsage >= plan.storage * 1.2 && "outOfStorage";             // 585
}                                                                                                            // 586
                                                                                                             // 587
isAdmin = function() {                                                                                       // 588
  // Returns true if the user is the administrator.                                                          // 589
                                                                                                             // 590
  var user = Meteor.user();                                                                                  // 591
  if (user && user.isAdmin) {                                                                                // 592
    return true;                                                                                             // 593
  } else {                                                                                                   // 594
    return false;                                                                                            // 595
  }                                                                                                          // 596
}                                                                                                            // 597
                                                                                                             // 598
isAdminById = function(id) {                                                                                 // 599
  // Returns true if the user's id is the administrator.                                                     // 600
                                                                                                             // 601
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                         // 602
  if (user && user.isAdmin) {                                                                                // 603
    return true;                                                                                             // 604
  } else {                                                                                                   // 605
    return false;                                                                                            // 606
  }                                                                                                          // 607
}                                                                                                            // 608
                                                                                                             // 609
findAdminUserForToken = function (token) {                                                                   // 610
  if (!token.requirements) {                                                                                 // 611
    return;                                                                                                  // 612
  }                                                                                                          // 613
  var requirements = token.requirements.filter(function (requirement) {                                      // 614
    return "userIsAdmin" in requirement;                                                                     // 615
  });                                                                                                        // 616
                                                                                                             // 617
  if (requirements.length > 1) {                                                                             // 618
    return;                                                                                                  // 619
  }                                                                                                          // 620
  if (requirements.length === 0) {                                                                           // 621
    return;                                                                                                  // 622
  }                                                                                                          // 623
  return requirements[0].userIsAdmin;                                                                        // 624
};                                                                                                           // 625
                                                                                                             // 626
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                             // 627
                                                                                                             // 628
if (wildcardHost.length != 2) {                                                                              // 629
  throw new Error("Wildcard host must contain exactly one asterisk.");                                       // 630
}                                                                                                            // 631
                                                                                                             // 632
matchWildcardHost = function(host) {                                                                         // 633
  // See if the hostname is a member of our wildcard. If so, extract the ID.                                 // 634
                                                                                                             // 635
  var prefix = wildcardHost[0];                                                                              // 636
  var suffix = wildcardHost[1];                                                                              // 637
                                                                                                             // 638
  // We remove everything after the first ":" character so that our                                          // 639
  // comparison logic ignores port numbers.                                                                  // 640
  suffix = suffix.split(":")[0];                                                                             // 641
  host = host.split(":")[0];                                                                                 // 642
                                                                                                             // 643
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                                    // 644
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                           // 645
      host.length >= prefix.length + suffix.length) {                                                        // 646
    var id = host.slice(prefix.length, -suffix.length);                                                      // 647
    if (id.match(/^[a-z0-9]*$/)) {                                                                           // 648
      return id;                                                                                             // 649
    }                                                                                                        // 650
  }                                                                                                          // 651
                                                                                                             // 652
  return null;                                                                                               // 653
}                                                                                                            // 654
                                                                                                             // 655
makeWildcardHost = function (id) {                                                                           // 656
  return wildcardHost[0] + id + wildcardHost[1];                                                             // 657
}                                                                                                            // 658
                                                                                                             // 659
if (Meteor.isServer) {                                                                                       // 660
  var Url = Npm.require("url");                                                                              // 661
  getWildcardOrigin = function () {                                                                          // 662
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're              // 663
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),            // 664
    // an astrisk is only allowed as the first character and must be followed by a period. So we need        // 665
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can           // 666
    // really do. We also add the protocol to the front (again, that's what CSP wants).                      // 667
                                                                                                             // 668
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.         // 669
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                                 // 670
                                                                                                             // 671
    var dotPos = wildcardHost[1].indexOf(".");                                                               // 672
    if (dotPos < 0) {                                                                                        // 673
      return protocol + "//*";                                                                               // 674
    } else {                                                                                                 // 675
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                               // 676
    }                                                                                                        // 677
  }                                                                                                          // 678
}                                                                                                            // 679
                                                                                                             // 680
allowDevAccounts = function () {                                                                             // 681
  var setting = Settings.findOne({_id: "devAccounts"});                                                      // 682
  if (setting) {                                                                                             // 683
    return setting.value;                                                                                    // 684
  } else {                                                                                                   // 685
    return Meteor.settings && Meteor.settings.public &&                                                      // 686
           Meteor.settings.public.allowDevAccounts;                                                          // 687
  }                                                                                                          // 688
};                                                                                                           // 689
                                                                                                             // 690
roleAssignmentPattern = {                                                                                    // 691
  none : Match.Optional(null),                                                                               // 692
  allAccess: Match.Optional(null),                                                                           // 693
  roleId: Match.Optional(Match.Integer),                                                                     // 694
  addPermissions: Match.Optional([Boolean]),                                                                 // 695
  removePermissions: Match.Optional([Boolean]),                                                              // 696
};                                                                                                           // 697
                                                                                                             // 698
SandstormDb = function () {                                                                                  // 699
  this.collections = {                                                                                       // 700
    // Direct access to underlying collections. DEPRECATED.                                                  // 701
    //                                                                                                       // 702
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove            // 703
    //   direct access to the collections.                                                                   // 704
                                                                                                             // 705
    packages: Packages,                                                                                      // 706
    devPackages: DevPackages,                                                                                // 707
    userActions: UserActions,                                                                                // 708
    grains: Grains,                                                                                          // 709
    contacts: Contacts,                                                                                      // 710
    sessions: Sessions,                                                                                      // 711
    signupKeys: SignupKeys,                                                                                  // 712
    activityStats: ActivityStats,                                                                            // 713
    deleteStats: DeleteStats,                                                                                // 714
    fileTokens: FileTokens,                                                                                  // 715
    apiTokens: ApiTokens,                                                                                    // 716
    notifications: Notifications,                                                                            // 717
    statsTokens: StatsTokens,                                                                                // 718
    misc: Misc,                                                                                              // 719
    settings: Settings,                                                                                      // 720
    appIndex: AppIndex,                                                                                      // 721
    keybaseProfiles: KeybaseProfiles,                                                                        // 722
                                                                                                             // 723
    // Intentionally omitted:                                                                                // 724
    // - Migrations, since it's used only within this package.                                               // 725
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.          // 726
  };                                                                                                         // 727
};                                                                                                           // 728
                                                                                                             // 729
// TODO(cleanup): These methods should not be defined freestanding and should use collection                 // 730
//   objects created in SandstormDb's constructor rather than globals.                                       // 731
                                                                                                             // 732
_.extend(SandstormDb.prototype, {                                                                            // 733
  isDemoUser: isDemoUser,                                                                                    // 734
  isSignedUp: isSignedUp,                                                                                    // 735
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                        // 736
  isUserOverQuota: isUserOverQuota,                                                                          // 737
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                                    // 738
  isAdmin: isAdmin,                                                                                          // 739
  isAdminById: isAdminById,                                                                                  // 740
  findAdminUserForToken: findAdminUserForToken,                                                              // 741
  matchWildcardHost: matchWildcardHost,                                                                      // 742
  makeWildcardHost: makeWildcardHost,                                                                        // 743
  allowDevAccounts: allowDevAccounts,                                                                        // 744
  roleAssignmentPattern: roleAssignmentPattern,                                                              // 745
});                                                                                                          // 746
                                                                                                             // 747
if (Meteor.isServer) {                                                                                       // 748
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                               // 749
}                                                                                                            // 750
                                                                                                             // 751
// =======================================================================================                   // 752
// Below this point are newly-written or refactored functions.                                               // 753
                                                                                                             // 754
_.extend(SandstormDb.prototype, {                                                                            // 755
  getUser: function getUser (userId) {                                                                       // 756
    check(userId, Match.OneOf(String, undefined, null));                                                     // 757
    if (userId) {                                                                                            // 758
      return Meteor.users.findOne(userId);                                                                   // 759
    }                                                                                                        // 760
  },                                                                                                         // 761
                                                                                                             // 762
  getIdentity: function getIdentity (identityId) {                                                           // 763
    check(identityId, String);                                                                               // 764
    // This would be a prime place to use Mongo's $ operator. Unfortunately, $ is not available              // 765
    // in Minimongo. Maybe this method should have separate server and client implementations?               // 766
    var user = Meteor.users.findOne({"identities.id": identityId});                                          // 767
    if (user) {                                                                                              // 768
      return _.findWhere(SandstormDb.getUserIdentities(user), {id: identityId});                             // 769
    }                                                                                                        // 770
  },                                                                                                         // 771
                                                                                                             // 772
  userHasIdentity: function (userId, identityId) {                                                           // 773
    check(userId, String);                                                                                   // 774
    check(identityId, String);                                                                               // 775
    return !!Meteor.users.findOne({_id: userId, "identities.id": identityId});                               // 776
  },                                                                                                         // 777
                                                                                                             // 778
  userGrains: function userGrains (userId) {                                                                 // 779
    check(userId, Match.OneOf(String, undefined, null));                                                     // 780
    return this.collections.grains.find({userId: userId});                                                   // 781
  },                                                                                                         // 782
                                                                                                             // 783
  currentUserGrains: function currentUserGrains () {                                                         // 784
    return this.userGrains(Meteor.userId());                                                                 // 785
  },                                                                                                         // 786
                                                                                                             // 787
  getGrain: function getGrain (grainId) {                                                                    // 788
    check(grainId, String);                                                                                  // 789
    return this.collections.grains.findOne(grainId);                                                         // 790
  },                                                                                                         // 791
                                                                                                             // 792
  userApiTokens: function userApiTokens (userId) {                                                           // 793
    check(userId, Match.OneOf(String, undefined, null));                                                     // 794
    identityIds = SandstormDb.getUserIdentities(this.getUser(userId))                                        // 795
        .map(function (identity) { return identity.id; });                                                   // 796
    return this.collections.apiTokens.find({'owner.user.identityId': {$in: identityIds}});                   // 797
  },                                                                                                         // 798
                                                                                                             // 799
  currentUserApiTokens: function currentUserApiTokens () {                                                   // 800
    return this.userApiTokens(Meteor.userId());                                                              // 801
  },                                                                                                         // 802
                                                                                                             // 803
  userActions: function userActions (user) {                                                                 // 804
    return this.collections.userActions.find({userId: user});                                                // 805
  },                                                                                                         // 806
                                                                                                             // 807
  currentUserActions: function currentUserActions () {                                                       // 808
    return this.userActions(Meteor.userId());                                                                // 809
  },                                                                                                         // 810
                                                                                                             // 811
  iconSrcForPackage: function iconSrcForPackage (pkg, usage) {                                               // 812
    return Identicon.iconSrcForPackage(pkg, usage, this.makeWildcardHost("static"));                         // 813
  },                                                                                                         // 814
                                                                                                             // 815
  getPlan: function (id) {                                                                                   // 816
    check(id, String);                                                                                       // 817
    var plan = Plans.findOne(id);                                                                            // 818
    if (!plan) {                                                                                             // 819
      throw new Error("no such plan: " + id);                                                                // 820
    }                                                                                                        // 821
    return plan;                                                                                             // 822
  },                                                                                                         // 823
                                                                                                             // 824
  listPlans: function () {                                                                                   // 825
    return Plans.find({}, {sort: {price: 1}});                                                               // 826
  },                                                                                                         // 827
                                                                                                             // 828
  getMyPlan: function () {                                                                                   // 829
    var user = Meteor.user();                                                                                // 830
    return user && Plans.findOne(user.plan || "free");                                                       // 831
  },                                                                                                         // 832
                                                                                                             // 833
  getMyUsage: function (user) {                                                                              // 834
    user = user || Meteor.user();                                                                            // 835
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                                     // 836
      if (Meteor.isClient) {                                                                                 // 837
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly              // 838
        // not reactive.                                                                                     // 839
        return user.pseudoUsage;                                                                             // 840
      } else {                                                                                               // 841
        return {                                                                                             // 842
          grains: Grains.find({userId: user._id}).count(),                                                   // 843
          storage: user.storageUsage || 0,                                                                   // 844
          compute: 0   // not tracked yet                                                                    // 845
        };                                                                                                   // 846
      }                                                                                                      // 847
    } else {                                                                                                 // 848
      return {grains: 0, storage: 0, compute: 0};                                                            // 849
    }                                                                                                        // 850
  },                                                                                                         // 851
                                                                                                             // 852
  isUninvitedFreeUser: function () {                                                                         // 853
    if (!Meteor.settings.public.allowUninvited) return false;                                                // 854
                                                                                                             // 855
    var user = Meteor.user();                                                                                // 856
    return user && !user.expires && (!user.plan || user.plan === "free");                                    // 857
  },                                                                                                         // 858
                                                                                                             // 859
  getSetting: function (name) {                                                                              // 860
    var setting = Settings.findOne(name);                                                                    // 861
    return setting && setting.value;                                                                         // 862
  },                                                                                                         // 863
                                                                                                             // 864
  addUserActions: function(packageId) {                                                                      // 865
    //TODO(cleanup): implement this with meteor methods rather than client-side inserts/removes.             // 866
    var pack = Packages.findOne(packageId);                                                                  // 867
    if (pack) {                                                                                              // 868
      // Remove old versions.                                                                                // 869
      UserActions.find({userId: Meteor.userId(), appId: pack.appId})                                         // 870
          .forEach(function (action) {                                                                       // 871
        UserActions.remove(action._id);                                                                      // 872
      });                                                                                                    // 873
                                                                                                             // 874
      // Install new.                                                                                        // 875
      var actions = pack.manifest.actions;                                                                   // 876
      for (i in actions) {                                                                                   // 877
        var action = actions[i];                                                                             // 878
        if ("none" in action.input) {                                                                        // 879
          var userAction = {                                                                                 // 880
            userId: Meteor.userId(),                                                                         // 881
            packageId: pack._id,                                                                             // 882
            appId: pack.appId,                                                                               // 883
            appTitle: pack.manifest.appTitle,                                                                // 884
            appMarketingVersion: pack.manifest.appMarketingVersion,                                          // 885
            appVersion: pack.manifest.appVersion,                                                            // 886
            title: action.title,                                                                             // 887
            nounPhrase: action.nounPhrase,                                                                   // 888
            command: action.command                                                                          // 889
          };                                                                                                 // 890
          UserActions.insert(userAction);                                                                    // 891
        } else {                                                                                             // 892
          // TODO(someday):  Implement actions with capability inputs.                                       // 893
        }                                                                                                    // 894
      }                                                                                                      // 895
                                                                                                             // 896
      Meteor.call("deleteUnusedPackages", pack.appId);                                                       // 897
    }                                                                                                        // 898
  },                                                                                                         // 899
                                                                                                             // 900
  sendAdminNotification: function (message, link) {                                                          // 901
    Meteor.users.find({isAdmin: true}, {fields: {_id: 1}}).forEach(function (user) {                         // 902
      Notifications.insert({                                                                                 // 903
        admin: {                                                                                             // 904
          action: link,                                                                                      // 905
          type: "reportStats",                                                                               // 906
        },                                                                                                   // 907
        userId: user._id,                                                                                    // 908
        text: {defaultText: message},                                                                        // 909
        timestamp: new Date(),                                                                               // 910
        isUnread: true,                                                                                      // 911
      });                                                                                                    // 912
    });                                                                                                      // 913
  },                                                                                                         // 914
                                                                                                             // 915
  getKeybaseProfile: function (keyFingerprint) {                                                             // 916
    return this.collections.keybaseProfiles.findOne(keyFingerprint) || {};                                   // 917
  },                                                                                                         // 918
});                                                                                                          // 919
                                                                                                             // 920
var appNameFromPackage = function(packageObj) {                                                              // 921
  // This function takes a Package object from Mongo and returns an                                          // 922
  // app title.                                                                                              // 923
  var manifest = packageObj.manifest;                                                                        // 924
  if (!manifest) return packageObj.appId || packageObj._id || "unknown";                                     // 925
  var action = manifest.actions[0];                                                                          // 926
  appName = (manifest.appTitle && manifest.appTitle.defaultText) ||                                          // 927
    appNameFromActionName(action.title.defaultText);                                                         // 928
  return appName;                                                                                            // 929
};                                                                                                           // 930
                                                                                                             // 931
var appNameFromActionName = function(name) {                                                                 // 932
  // Hack: Historically we only had action titles, like "New Etherpad Document", not app                     // 933
  //   titles. But for this UI we want app titles. As a transitionary measure, try to                        // 934
  //   derive the app title from the action title.                                                           // 935
  // TODO(cleanup): Get rid of this once apps have real titles.                                              // 936
  if (!name) {                                                                                               // 937
    return "(unnamed)";                                                                                      // 938
  }                                                                                                          // 939
  if (name.lastIndexOf("New ", 0) === 0) {                                                                   // 940
    name = name.slice(4);                                                                                    // 941
  }                                                                                                          // 942
  if (name.lastIndexOf("Hacker CMS", 0) === 0) {                                                             // 943
    name = "Hacker CMS";                                                                                     // 944
  } else {                                                                                                   // 945
    var space = name.indexOf(" ");                                                                           // 946
    if (space > 0) {                                                                                         // 947
      name = name.slice(0, space);                                                                           // 948
    }                                                                                                        // 949
  }                                                                                                          // 950
  return name;                                                                                               // 951
};                                                                                                           // 952
                                                                                                             // 953
var appShortDescriptionFromPackage = function (pkg) {                                                        // 954
  return pkg && pkg.manifest && pkg.manifest.metadata &&                                                     // 955
         pkg.manifest.metadata.shortDescription &&                                                           // 956
         pkg.manifest.metadata.shortDescription.defaultText;                                                 // 957
};                                                                                                           // 958
                                                                                                             // 959
var nounPhraseForActionAndAppTitle = function(action, appTitle) {                                            // 960
  // A hack to deal with legacy apps not including fields in their manifests.                                // 961
  // I look forward to the day I can remove most of this code.                                               // 962
  // Attempt to figure out the appropriate noun that this action will create.                                // 963
  // Use an explicit noun phrase is one is available.  Apps should add these in the future.                  // 964
  if (action.nounPhrase) return action.nounPhrase.defaultText;                                               // 965
  // Otherwise, try to guess one from the structure of the action title field                                // 966
  if (action.title && action.title.defaultText) {                                                            // 967
    var text = action.title.defaultText;                                                                     // 968
    // Strip a leading "New "                                                                                // 969
    if (text.lastIndexOf("New ", 0) === 0) {                                                                 // 970
      var candidate = text.slice(4);                                                                         // 971
      // Strip a leading appname too, if provided                                                            // 972
      if (candidate.lastIndexOf(appTitle, 0) === 0) {                                                        // 973
        var newCandidate = candidate.slice(appTitle.length);                                                 // 974
        // Unless that leaves you with no noun, in which case, use "grain"                                   // 975
        if (newCandidate.length > 0) {                                                                       // 976
          return newCandidate.toLowerCase();                                                                 // 977
        } else {                                                                                             // 978
          return "grain";                                                                                    // 979
        }                                                                                                    // 980
      }                                                                                                      // 981
      return candidate.toLowerCase();                                                                        // 982
    }                                                                                                        // 983
    // Some other verb phrase was given.  Just use it verbatim, and hope the app author updates              // 984
    // the package soon.                                                                                     // 985
    return text;                                                                                             // 986
  } else {                                                                                                   // 987
    return "grain";                                                                                          // 988
  }                                                                                                          // 989
};                                                                                                           // 990
                                                                                                             // 991
// Static methods on SandstormDb that don't need an instance.                                                // 992
// Largely things that deal with backwards-compatibility.                                                    // 993
_.extend(SandstormDb, {                                                                                      // 994
  appNameFromActionName: appNameFromActionName,                                                              // 995
  appNameFromPackage: appNameFromPackage,                                                                    // 996
  appShortDescriptionFromPackage: appShortDescriptionFromPackage,                                            // 997
  nounPhraseForActionAndAppTitle: nounPhraseForActionAndAppTitle,                                            // 998
});                                                                                                          // 999
                                                                                                             // 1000
if (Meteor.isServer) {                                                                                       // 1001
  var Crypto = Npm.require("crypto");                                                                        // 1002
  var ContentType = Npm.require("content-type");                                                             // 1003
  var Zlib = Npm.require("zlib");                                                                            // 1004
                                                                                                             // 1005
  var replicaNumber = Meteor.settings.replicaNumber || 0;                                                    // 1006
                                                                                                             // 1007
  var computeStagger = function (n) {                                                                        // 1008
    // Compute a fraction in the range [0, 1) such that, for any natural number k, the values                // 1009
    // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.                 // 1010
    // The sequence looks like:                                                                              // 1011
    //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                                     // 1012
    //                                                                                                       // 1013
    // We use this to determine how we'll stagger periodic events performed by this replica.                 // 1014
    // Notice that this allows us to compute a stagger which is independent of the number of                 // 1015
    // front-end replicas present; we can add more replicas to the end without affecting how the             // 1016
    // earlier ones schedule their events.                                                                   // 1017
    var denom = 1;                                                                                           // 1018
    while (denom <= n) denom <<= 1;                                                                          // 1019
    var num = n * 2 - denom + 1;                                                                             // 1020
    return num / denom;                                                                                      // 1021
  }                                                                                                          // 1022
                                                                                                             // 1023
  var stagger = computeStagger(replicaNumber);                                                               // 1024
                                                                                                             // 1025
  SandstormDb.periodicCleanup = function (intervalMs, callback) {                                            // 1026
    // Register a database cleanup function than should run periodically, roughly once every                 // 1027
    // interval of the given length.                                                                         // 1028
    //                                                                                                       // 1029
    // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be              // 1030
    // scaled appropriately on the assumption that more data is being generated demanding more               // 1031
    // frequent cleanups.                                                                                    // 1032
                                                                                                             // 1033
    check(intervalMs, Number);                                                                               // 1034
    check(callback, Function);                                                                               // 1035
                                                                                                             // 1036
    if (intervalMs < 120000) {                                                                               // 1037
      throw new Error("less than 2-minute cleanup interval seems too fast; " +                               // 1038
                      "are you using the right units?");                                                     // 1039
    }                                                                                                        // 1040
                                                                                                             // 1041
    // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that              // 1042
    // the schedule is independent of the exact startup time.                                                // 1043
    var first = intervalMs - Date.now() % intervalMs;                                                        // 1044
                                                                                                             // 1045
    // Stagger cleanups across replicas so that we don't have all replicas trying to clean the               // 1046
    // same data at the same time.                                                                           // 1047
    first += Math.floor(intervalMs * computeStagger(replicaNumber));                                         // 1048
                                                                                                             // 1049
    // If the stagger put us more than an interval away from now, back up.                                   // 1050
    if (first > intervalMs) first -= intervalMs;                                                             // 1051
                                                                                                             // 1052
    Meteor.setTimeout(function () {                                                                          // 1053
      callback();                                                                                            // 1054
      Meteor.setInterval(callback, intervalMs);                                                              // 1055
    }, first);                                                                                               // 1056
  }                                                                                                          // 1057
                                                                                                             // 1058
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.                  // 1059
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                          // 1060
                                                                                                             // 1061
  var BufferSmallerThan = function (limit) {                                                                 // 1062
    return Match.Where(function (buf) {                                                                      // 1063
      check(buf, Buffer);                                                                                    // 1064
      return buf.length < limit;                                                                             // 1065
    });                                                                                                      // 1066
  }                                                                                                          // 1067
                                                                                                             // 1068
  var DatabaseId = Match.Where(function (s) {                                                                // 1069
    check(s, String);                                                                                        // 1070
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                                     // 1071
  });                                                                                                        // 1072
                                                                                                             // 1073
  addStaticAsset = function (metadata, content) {                                                            // 1074
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it             // 1075
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.          // 1076
                                                                                                             // 1077
    if (typeof content === "string" && !metadata.encoding) {                                                 // 1078
      content = gzipSync(new Buffer(content, "utf8"));                                                       // 1079
      metadata.encoding = "gzip";                                                                            // 1080
    }                                                                                                        // 1081
                                                                                                             // 1082
    check(metadata, {                                                                                        // 1083
      mimeType: String,                                                                                      // 1084
      encoding: Match.Optional("gzip")                                                                       // 1085
    });                                                                                                      // 1086
    check(content, BufferSmallerThan(1 << 20));                                                              // 1087
                                                                                                             // 1088
    // Validate content type.                                                                                // 1089
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                            // 1090
                                                                                                             // 1091
    var hasher = Crypto.createHash("sha256");                                                                // 1092
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                              // 1093
    hasher.update(content);                                                                                  // 1094
    var hash = hasher.digest("base64");                                                                      // 1095
                                                                                                             // 1096
    var existing = StaticAssets.findAndModify({                                                              // 1097
      query: {hash: hash, refcount: {$gte: 1}},                                                              // 1098
      update: {$inc: {refcount: 1}},                                                                         // 1099
      fields: {_id: 1, refcount: 1},                                                                         // 1100
    });                                                                                                      // 1101
    if (existing) {                                                                                          // 1102
      return existing._id;                                                                                   // 1103
    }                                                                                                        // 1104
                                                                                                             // 1105
    return StaticAssets.insert(_.extend({                                                                    // 1106
      hash: hash,                                                                                            // 1107
      content: content,                                                                                      // 1108
      refcount: 1                                                                                            // 1109
    }, metadata));                                                                                           // 1110
  }                                                                                                          // 1111
                                                                                                             // 1112
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                                     // 1113
                                                                                                             // 1114
  SandstormDb.prototype.refStaticAsset = function (id) {                                                     // 1115
    // Increment the refcount on an existing static asset.                                                   // 1116
    //                                                                                                       // 1117
    // You must call this BEFORE adding the new reference to the DB, in case of failure between              // 1118
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1119
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1120
                                                                                                             // 1121
    check(id, String);                                                                                       // 1122
                                                                                                             // 1123
    var existing = StaticAssets.findAndModify({                                                              // 1124
      query: {hash: hash},                                                                                   // 1125
      update: {$inc: {refcount: 1}},                                                                         // 1126
      fields: {_id: 1, refcount: 1},                                                                         // 1127
    });                                                                                                      // 1128
    if (!existing) {                                                                                         // 1129
      throw new Error("refStaticAsset() called on asset that doesn't exist");                                // 1130
    }                                                                                                        // 1131
  }                                                                                                          // 1132
                                                                                                             // 1133
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                                   // 1134
    // Decrement refcount on a static asset and delete if it has reached zero.                               // 1135
    //                                                                                                       // 1136
    // You must call this AFTER removing the reference from the DB, in case of failure between               // 1137
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1138
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1139
                                                                                                             // 1140
    check(id, String);                                                                                       // 1141
                                                                                                             // 1142
    var existing = StaticAssets.findAndModify({                                                              // 1143
      query: {_id: id},                                                                                      // 1144
      update: {$inc: {refcount: -1}},                                                                        // 1145
      fields: {_id: 1, refcount: 1},                                                                         // 1146
      new: true,                                                                                             // 1147
    });                                                                                                      // 1148
    if (!existing) {                                                                                         // 1149
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);               // 1150
    } else if (existing.refcount <= 0) {                                                                     // 1151
      StaticAssets.remove({_id: existing._id});                                                              // 1152
    }                                                                                                        // 1153
  }                                                                                                          // 1154
                                                                                                             // 1155
  SandstormDb.prototype.getStaticAsset = function (id) {                                                     // 1156
    // Get a static asset's mimeType, encoding, and raw content.                                             // 1157
                                                                                                             // 1158
    check(id, String);                                                                                       // 1159
                                                                                                             // 1160
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});          // 1161
    if (asset) {                                                                                             // 1162
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy              // 1163
      //   here.                                                                                             // 1164
      asset.content = new Buffer(asset.content);                                                             // 1165
    }                                                                                                        // 1166
    return asset;                                                                                            // 1167
  }                                                                                                          // 1168
                                                                                                             // 1169
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                                // 1170
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: DatabaseId}});                          // 1171
                                                                                                             // 1172
    return AssetUploadTokens.insert({                                                                        // 1173
      purpose: purpose,                                                                                      // 1174
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                               // 1175
    });                                                                                                      // 1176
  }                                                                                                          // 1177
                                                                                                             // 1178
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                                 // 1179
    // Indicates that the given asset upload has completed. It will be removed and its purpose               // 1180
    // returned. If no matching upload exists, returns undefined.                                            // 1181
                                                                                                             // 1182
    check(id, String);                                                                                       // 1183
                                                                                                             // 1184
    var upload = AssetUploadTokens.findAndModify({                                                           // 1185
      query: {_id: id},                                                                                      // 1186
      remove: true                                                                                           // 1187
    });                                                                                                      // 1188
                                                                                                             // 1189
    if (upload.expires.valueOf() < Date.now()) {                                                             // 1190
      return undefined;  // already expired                                                                  // 1191
    } else {                                                                                                 // 1192
      return upload.purpose;                                                                                 // 1193
    }                                                                                                        // 1194
  }                                                                                                          // 1195
                                                                                                             // 1196
  function cleanupExpiredAssetUploads() {                                                                    // 1197
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                                  // 1198
  }                                                                                                          // 1199
                                                                                                             // 1200
  // Cleanup tokens every hour.                                                                              // 1201
  SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                          // 1202
                                                                                                             // 1203
  var packageCache = {};                                                                                     // 1204
  // Package info is immutable. Let's cache to save on mongo queries.                                        // 1205
                                                                                                             // 1206
  SandstormDb.prototype.getPackage = function (packageId) {                                                  // 1207
    // Get the given package record. Since package info is immutable, cache the data in the server           // 1208
    // to reduce mongo query overhead, since it turns out we have to fetch specific packages a               // 1209
    // lot.                                                                                                  // 1210
                                                                                                             // 1211
    if (packageId in packageCache) {                                                                         // 1212
      return packageCache[packageId];                                                                        // 1213
    }                                                                                                        // 1214
                                                                                                             // 1215
    var package = Packages.findOne(packageId);                                                               // 1216
    if (package && package.status === "ready") {                                                             // 1217
      packageCache[packageId] = package;                                                                     // 1218
    }                                                                                                        // 1219
    return package;                                                                                          // 1220
  }                                                                                                          // 1221
                                                                                                             // 1222
  SandstormDb.prototype.sendAppUpdateNotifications = function (appId, packageId, name,                       // 1223
                                                               versionNumber, marketingVersion) {            // 1224
    var db = this;                                                                                           // 1225
    var actions = db.collections.userActions.find({appId: appId, appVersion: {$lt: versionNumber}},          // 1226
      {fields: {userId: 1}});                                                                                // 1227
    actions.forEach(function (action) {                                                                      // 1228
      var userId = action.userId;                                                                            // 1229
      var updater = {                                                                                        // 1230
        userId: userId,                                                                                      // 1231
        timestamp: new Date(),                                                                               // 1232
        isUnread: true,                                                                                      // 1233
      };                                                                                                     // 1234
                                                                                                             // 1235
      // Set only the appId that we care about. Use mongo's dot notation to specify only a single            // 1236
      // field inside of an object to update                                                                 // 1237
      updater["appUpdates." + appId] = {                                                                     // 1238
        marketingVersion: marketingVersion,                                                                  // 1239
        packageId: packageId,                                                                                // 1240
        name: name,                                                                                          // 1241
        version: versionNumber,                                                                              // 1242
      };                                                                                                     // 1243
      db.collections.notifications.upsert({userId: userId}, { $set: updater});                               // 1244
    });                                                                                                      // 1245
                                                                                                             // 1246
    db.collections.appIndex.update({_id: appId}, {$set: {hasSentNotifications: true}});                      // 1247
                                                                                                             // 1248
    // In the case where we replaced a previous notification and that was the only reference to the          // 1249
    // package, we need to clean it up                                                                       // 1250
    Meteor.call("deleteUnusedPackages", appId);                                                              // 1251
  };                                                                                                         // 1252
                                                                                                             // 1253
  SandstormDb.prototype.upgradeGrains =  function (appId, version, packageId, backend) {                     // 1254
    check(appId, String);                                                                                    // 1255
    check(version, Match.Integer);                                                                           // 1256
    check(packageId, String);                                                                                // 1257
                                                                                                             // 1258
    var selector = {                                                                                         // 1259
      userId: Meteor.userId(),                                                                               // 1260
      appId: appId,                                                                                          // 1261
      appVersion: { $lte: version },                                                                         // 1262
      packageId: { $ne: packageId }                                                                          // 1263
    };                                                                                                       // 1264
                                                                                                             // 1265
    if (!this.isSimulation) {                                                                                // 1266
      Grains.find(selector).forEach(function (grain) {                                                       // 1267
        backend.shutdownGrain(grain._id, grain.userId);                                                      // 1268
      });                                                                                                    // 1269
    }                                                                                                        // 1270
                                                                                                             // 1271
    Grains.update(selector, { $set: { appVersion: version, packageId: packageId }}, {multi: true});          // 1272
  };                                                                                                         // 1273
                                                                                                             // 1274
  SandstormDb.prototype.startInstall = function (packageId, url, retryFailed, isAutoUpdated) {               // 1275
    // Mark package for possible installation.                                                               // 1276
                                                                                                             // 1277
    var fields = {status: "download", progress: 0, url: url, isAutoUpdated: !!isAutoUpdated};                // 1278
                                                                                                             // 1279
    if (retryFailed) {                                                                                       // 1280
      Packages.update({_id: packageId, status: "failed"}, {$set: fields});                                   // 1281
    } else {                                                                                                 // 1282
      try {                                                                                                  // 1283
        fields._id = packageId;                                                                              // 1284
        Packages.insert(fields);                                                                             // 1285
      } catch (err) {                                                                                        // 1286
        console.error("Simultaneous startInstall()s?", err.stack);                                           // 1287
      }                                                                                                      // 1288
    }                                                                                                        // 1289
  };                                                                                                         // 1290
                                                                                                             // 1291
  var ValidKeyFingerprint = Match.Where(function (keyFingerprint) {                                          // 1292
    check(keyFingerprint, String);                                                                           // 1293
    return !!keyFingerprint.match(/^[0-9A-F]{40}$/);                                                         // 1294
  });                                                                                                        // 1295
                                                                                                             // 1296
  SandstormDb.prototype.updateKeybaseProfileAsync = function (keyFingerprint) {                              // 1297
    // Asynchronously fetch the given Keybase profile and populate the KeybaseProfiles collection.           // 1298
                                                                                                             // 1299
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1300
                                                                                                             // 1301
    console.log("fetching keybase", keyFingerprint);                                                         // 1302
                                                                                                             // 1303
    HTTP.get(                                                                                                // 1304
        "https://keybase.io/_/api/1.0/user/lookup.json?key_fingerprint=" + keyFingerprint +                  // 1305
        "&fields=basics,profile,proofs_summary", {                                                           // 1306
      timeout: 5000                                                                                          // 1307
    }, function (err, keybaseResponse) {                                                                     // 1308
      if (err) {                                                                                             // 1309
        console.log("keybase lookup error:", err.stack);                                                     // 1310
        return;                                                                                              // 1311
      }                                                                                                      // 1312
                                                                                                             // 1313
      if (!keybaseResponse.data) {                                                                           // 1314
        console.log("keybase didn't return JSON? Headers:", keybaseResponse.headers);                        // 1315
        return;                                                                                              // 1316
      }                                                                                                      // 1317
                                                                                                             // 1318
      var profile = (keybaseResponse.data.them || [])[0];                                                    // 1319
                                                                                                             // 1320
      if (profile) {                                                                                         // 1321
        var record = {                                                                                       // 1322
          displayName: (profile.profile || {}).full_name,                                                    // 1323
          handle: (profile.basics || {}).username,                                                           // 1324
          proofs: (profile.proofs_summary || {}).all || []                                                   // 1325
        };                                                                                                   // 1326
                                                                                                             // 1327
        record.proofs.forEach(function (proof) {                                                             // 1328
          // Remove potentially Mongo-incompatible stuff. (Currently Keybase returns nothing that            // 1329
          // this would filter.)                                                                             // 1330
          for (var field in proof) {                                                                         // 1331
            // Don't allow field names containing '.' or '$'. Also don't allow sub-objects mainly            // 1332
            // because I'm too lazy to check the field names recursively (and Keybase doesn't                // 1333
            // return any objects anyway).                                                                   // 1334
            if (field.match(/[.$]/) || typeof(proof[field]) === "object") {                                  // 1335
              delete proof[field];                                                                           // 1336
            }                                                                                                // 1337
          }                                                                                                  // 1338
                                                                                                             // 1339
          // Indicate not verified.                                                                          // 1340
          // TODO(security): Asynchronously verify proofs. Presumably we can borrow code from the            // 1341
          //   Keybase node-based CLI.                                                                       // 1342
          proof.status = "unverified";                                                                       // 1343
        });                                                                                                  // 1344
                                                                                                             // 1345
        KeybaseProfiles.update(keyFingerprint, {$set: record}, {upsert: true});                              // 1346
      } else {                                                                                               // 1347
        // Keybase reports no match, so remove what we know of this user. We don't want to remove            // 1348
        // the item entirely from the cache as this will cause us to repeatedly re-fetch the data            // 1349
        // from Keybase.                                                                                     // 1350
        //                                                                                                   // 1351
        // TODO(someday): We could perhaps keep the proofs if we can still verify them directly,             // 1352
        //   but at present we don't have the ability to verify proofs.                                      // 1353
        KeybaseProfiles.update(keyFingerprint,                                                               // 1354
            {$unset: {displayName: "", handle: "", proofs: ""}}, {upsert: true});                            // 1355
      }                                                                                                      // 1356
    });                                                                                                      // 1357
  };                                                                                                         // 1358
                                                                                                             // 1359
  Meteor.publish("keybaseProfile", function (keyFingerprint) {                                               // 1360
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1361
    var db = this.connection.sandstormDb;                                                                    // 1362
                                                                                                             // 1363
    var cursor = db.collections.keybaseProfiles.find(keyFingerprint);                                        // 1364
    if (cursor.count() === 0) {                                                                              // 1365
      // Fire off async update.                                                                              // 1366
      db.updateKeybaseProfileAsync(keyFingerprint);                                                          // 1367
    }                                                                                                        // 1368
                                                                                                             // 1369
    return cursor;                                                                                           // 1370
  });                                                                                                        // 1371
                                                                                                             // 1372
  Meteor.publish("appIndex", function (appId) {                                                              // 1373
    check(appId, String);                                                                                    // 1374
    var db = this.connection.sandstormDb;                                                                    // 1375
    var cursor = db.collections.appIndex.find({_id: appId});                                                 // 1376
    return cursor;                                                                                           // 1377
  });                                                                                                        // 1378
                                                                                                             // 1379
  Meteor.publish("userPackages", function() {                                                                // 1380
    // Users should be able to see packages that are either:                                                 // 1381
    // 1. referenced by one of their userActions                                                             // 1382
    // 2. referenced by one of their grains                                                                  // 1383
    var self = this;                                                                                         // 1384
    var db = this.connection.sandstormDb;                                                                    // 1385
                                                                                                             // 1386
    // Note that package information, once it is in the database, is static. There's no need to              // 1387
    // reactively subscribe to changes to a package since they don't change. It's also unecessary            // 1388
    // to reactively remove a package from the client side when it is removed on the server, or              // 1389
    // when the client stops using it, because the worst case is the client has a small amount               // 1390
    // of extra info on a no-longer-used package held in memory until they refresh Sandstorm.                // 1391
    // So, we implement this as a cache: the first time each package ID shows up among the user's            // 1392
    // stuff, we push the package info to the client, and then we never update it.                           // 1393
    //                                                                                                       // 1394
    // Alternatively, we could subscribe to each individual package query, but this would waste              // 1395
    // lots of server-side resources watching for events that will never happen or don't matter.             // 1396
    var hasPackage = {};                                                                                     // 1397
    var refPackage = function (packageId) {                                                                  // 1398
      // Ignore dev apps.                                                                                    // 1399
      if (packageId.lastIndexOf("dev-", 0) === 0) return;                                                    // 1400
                                                                                                             // 1401
      if (!hasPackage[packageId]) {                                                                          // 1402
        hasPackage[packageId] = true;                                                                        // 1403
        var pkg = db.getPackage(packageId);                                                                  // 1404
        if (pkg) {                                                                                           // 1405
          self.added("packages", packageId, pkg);                                                            // 1406
        } else {                                                                                             // 1407
          console.error(                                                                                     // 1408
              "shouldn't happen: missing package referenced by user's stuff:", packageId);                   // 1409
        }                                                                                                    // 1410
      }                                                                                                      // 1411
    };                                                                                                       // 1412
                                                                                                             // 1413
    // package source 1: packages referred to by actions                                                     // 1414
    var actions = db.userActions(this.userId);                                                               // 1415
    var actionsHandle = actions.observe({                                                                    // 1416
      added: function(newAction) {                                                                           // 1417
        refPackage(newAction.packageId);                                                                     // 1418
      },                                                                                                     // 1419
      changed: function(oldAction, newAction) {                                                              // 1420
        refPackage(newAction.packageId);                                                                     // 1421
      }                                                                                                      // 1422
    });                                                                                                      // 1423
                                                                                                             // 1424
    // package source 2: packages referred to by grains directly                                             // 1425
    var grains = db.userGrains(this.userId);                                                                 // 1426
    var grainsHandle = grains.observe({                                                                      // 1427
      added: function(newGrain) {                                                                            // 1428
        refPackage(newGrain.packageId);                                                                      // 1429
      },                                                                                                     // 1430
      changed: function(oldGrain, newGrain) {                                                                // 1431
        refPackage(newGrain.packageId);                                                                      // 1432
      }                                                                                                      // 1433
    });                                                                                                      // 1434
                                                                                                             // 1435
    this.onStop(function () {                                                                                // 1436
      actionsHandle.stop();                                                                                  // 1437
      grainsHandle.stop();                                                                                   // 1438
    });                                                                                                      // 1439
    this.ready();                                                                                            // 1440
  });                                                                                                        // 1441
}                                                                                                            // 1442
                                                                                                             // 1443
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/profile.js                                                                          //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var makeIdenticon;                                                                                           // 17
var httpProtocol;                                                                                            // 18
                                                                                                             // 19
if (Meteor.isServer) {                                                                                       // 20
  Meteor.publish("accountIdentities", function () {                                                          // 21
    if (!this.userId) return [];                                                                             // 22
                                                                                                             // 23
    return [                                                                                                 // 24
      Meteor.users.find(this.userId,                                                                         // 25
        {fields: {                                                                                           // 26
          "identities.id":1,                                                                                 // 27
          "identities.service":1,                                                                            // 28
          "identities.profile":1,                                                                            // 29
          "identities.verifiedEmail":1,                                                                      // 30
          "identities.unverifiedEmail":1,                                                                    // 31
          "identities.main":1,                                                                               // 32
          "identities.noLogin":1,                                                                            // 33
                                                                                                             // 34
          "expires":1,                                                                                       // 35
                                                                                                             // 36
          "services.google.id":1,                                                                            // 37
          "services.google.email":1,                                                                         // 38
          "services.google.verified_email":1,                                                                // 39
          "services.google.name":1,                                                                          // 40
          "services.google.picture":1,                                                                       // 41
          "services.google.gender":1,                                                                        // 42
                                                                                                             // 43
          "services.github.id":1,                                                                            // 44
          "services.github.email":1,                                                                         // 45
          "services.github.username":1,                                                                      // 46
        }})                                                                                                  // 47
    ];                                                                                                       // 48
  });                                                                                                        // 49
                                                                                                             // 50
  makeIdenticon = function (id) {                                                                            // 51
    // We only make identicons client-side.                                                                  // 52
    return undefined;                                                                                        // 53
  }                                                                                                          // 54
                                                                                                             // 55
  var Url = Npm.require("url");                                                                              // 56
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                                   // 57
} else {                                                                                                     // 58
  var identiconCache = {};                                                                                   // 59
                                                                                                             // 60
  makeIdenticon = function (id) {                                                                            // 61
    // Given a cryptographic hash as input, generate an identicon. We always use the user's                  // 62
    // identity ID as the input hash (the same thing passed to apps as the user ID).                         // 63
                                                                                                             // 64
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so               // 65
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out           // 66
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it              // 67
    // anyway to be safe.                                                                                    // 68
    id = id.slice(0, 32);                                                                                    // 69
                                                                                                             // 70
    if (id in identiconCache) {                                                                              // 71
      return identiconCache[id];                                                                             // 72
    }                                                                                                        // 73
                                                                                                             // 74
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.             // 75
    var data = new Identicon(id, 512).toString();                                                            // 76
    var result = "data:image/png;base64," + data;                                                            // 77
    identiconCache[id] = result;                                                                             // 78
    return result;                                                                                           // 79
  }                                                                                                          // 80
                                                                                                             // 81
  httpProtocol = window.location.protocol;                                                                   // 82
}                                                                                                            // 83
                                                                                                             // 84
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                          // 85
                                                                                                             // 86
function staticAssetUrl(id, staticHost) {                                                                    // 87
  if (id) {                                                                                                  // 88
    return staticHost + "/" + id;                                                                            // 89
  } else {                                                                                                   // 90
    return undefined;                                                                                        // 91
  }                                                                                                          // 92
}                                                                                                            // 93
                                                                                                             // 94
function filterHandle(handle) {                                                                              // 95
  // Convert disallowed letters into underscores, but no more than one underscore in a row.                  // 96
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                                     // 97
}                                                                                                            // 98
                                                                                                             // 99
function emailToHandle(email) {                                                                              // 100
  // Turn an email address into a handle.                                                                    // 101
                                                                                                             // 102
  if (!email) return undefined;                                                                              // 103
                                                                                                             // 104
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters               // 105
  // on the same account.                                                                                    // 106
  var parts = email.split("@");                                                                              // 107
  var base = filterHandle(parts[0].split("+")[0]);                                                           // 108
                                                                                                             // 109
  var domain = (parts[1]||"").split(".");                                                                    // 110
  if (domain[domain.length - 1] === "name") {                                                                // 111
    // Oh, a .name domain. Let's use                                                                         // 112
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                       // 113
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",                  // 114
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],                 // 115
                         base)) {                                                                            // 116
    // This is probably an address at a vanity domain. Use the domain itself as the handle.                  // 117
    base = filterHandle(domain[0]);                                                                          // 118
  }                                                                                                          // 119
                                                                                                             // 120
  return filterHandle(base);                                                                                 // 121
}                                                                                                            // 122
                                                                                                             // 123
function fillInDefaults(identity, user) {                                                                    // 124
  if (identity.service.github) {                                                                             // 125
    identity.profile.name = identity.profile.name || user.services.github.username || "Name Unknown";        // 126
    identity.profile.handle = identity.profile.handle ||                                                     // 127
        filterHandle(user.services.github.username) ||                                                       // 128
        filterHandle(identity.profile.name) || "unknown";                                                    // 129
  } else if (identity.service.google) {                                                                      // 130
    identity.profile.name = identity.profile.name || user.services.google.name || "Name Unknown";            // 131
    identity.profile.handle = identity.profile.handle || emailToHandle(user.services.google.email) ||        // 132
        filterHandle(identity.profile.name) || "unknown";                                                    // 133
    identity.profile.pronoun = identity.profile.pronoun || GENDERS[user.services.google.gender] ||           // 134
        "neutral";                                                                                           // 135
  } else if (identity.service.email) {                                                                       // 136
    identity.profile.name = identity.profile.name || identity.verifiedEmail.split("@")[0];                   // 137
    identity.profile.handle = identity.profile.handle || emailToHandle(identity.verifiedEmail);              // 138
  } else if (identity.service.dev) {                                                                         // 139
    var lowerCaseName = identity.service.dev.name.split(" ")[0].toLowerCase();                               // 140
    identity.profile.name = identity.profile.name || identity.service.dev.name;                              // 141
    identity.profile.handle = identity.profile.handle || filterHandle(lowerCaseName);                        // 142
    identity.profile.pronoun = identity.profile.pronoun ||                                                   // 143
        (_.contains(["alice", "carol", "eve"], lowerCaseName) ? "female" :                                   // 144
         _.contains(["bob", "dave"], lowerCaseName) ? "male" : "neutral");                                   // 145
  } else if (identity.service.demo) {                                                                        // 146
    identity.profile.name = identity.profile.name || "Demo User";                                            // 147
    identity.profile.handle = identity.profile.handle || "demo";                                             // 148
  } else {                                                                                                   // 149
    throw new Error("unrecognized identity service: ", identity.service);                                    // 150
  }                                                                                                          // 151
                                                                                                             // 152
  identity.pronoun = identity.pronoun || "netural";                                                          // 153
}                                                                                                            // 154
                                                                                                             // 155
SandstormDb.getUserIdentities = function (user) {                                                            // 156
  // Given a user object, return all of the user's identities.                                               // 157
  //                                                                                                         // 158
  // On the client, must be subscribed "accountIdentities" for the user.                                     // 159
  if (!user || !user.identities) return [];                                                                  // 160
                                                                                                             // 161
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                         // 162
  return user.identities.map(function(identity) {                                                            // 163
    identity.profile.pictureUrl = staticAssetUrl(identity.profile.picture, staticHost) ||                    // 164
        makeIdenticon(identity.id);                                                                          // 165
    fillInDefaults(identity, user);                                                                          // 166
    return identity;                                                                                         // 167
  });                                                                                                        // 168
}                                                                                                            // 169
                                                                                                             // 170
SandstormDb.getUserEmails = function (user) {                                                                // 171
  // Given a user object, returns an array containing all email addresses associated with that user.         // 172
  // Each entry in the array is an object of the form:                                                       // 173
  //     `{email: String, verified: Bool, primary: Optional(Bool)}`                                          // 174
  //                                                                                                         // 175
  // At most one entry in the result has `primary = true`.                                                   // 176
                                                                                                             // 177
  var result = [];                                                                                           // 178
  if (!user || !user.identities) return result;                                                              // 179
                                                                                                             // 180
  user.identities.forEach(function (identity) {                                                              // 181
    if (identity.verifiedEmail) {                                                                            // 182
      result.push({email: identity.verifiedEmail, verified: true});                                          // 183
    }                                                                                                        // 184
    if (identity.unverifiedEmail) {                                                                          // 185
      result.push({email: identity.unverifiedEmail, verified: false});                                       // 186
    }                                                                                                        // 187
  });                                                                                                        // 188
                                                                                                             // 189
  // TODO(soon): Allow the user to select a verified email as their primary email.                           // 190
  for (var ii = 0; ii < result.length; ++ii) {                                                               // 191
    if (result[ii].verified) {                                                                               // 192
      result[ii].primary = true;                                                                             // 193
      break;                                                                                                 // 194
    } else if (ii == result.length - 1) {                                                                    // 195
      // No verified addresses. Mark the first address as primary.                                           // 196
      result[0].primary = true;                                                                              // 197
    }                                                                                                        // 198
  }                                                                                                          // 199
  return result;                                                                                             // 200
}                                                                                                            // 201
                                                                                                             // 202
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/user.js                                                                             //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var Crypto = Npm.require("crypto");                                                                          // 17
var Future = Npm.require("fibers/future");                                                                   // 18
                                                                                                             // 19
userPictureUrl = function (user) {                                                                           // 20
  if (user.services && !(user.profile && user.profile.picture)) {                                            // 21
    // Try to determine user's avatar URL from login service.                                                // 22
                                                                                                             // 23
    var google = user.services.google;                                                                       // 24
    if (google && google.picture) {                                                                          // 25
      return google.picture;                                                                                 // 26
    }                                                                                                        // 27
                                                                                                             // 28
    var github = user.services.github;                                                                       // 29
    if (github && github.id) {                                                                               // 30
      return "https://avatars.githubusercontent.com/u/" + github.id;                                         // 31
    }                                                                                                        // 32
                                                                                                             // 33
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be            // 34
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                // 35
    // Github are different because they are actually the identity providers, so they already know           // 36
    // the user logged in.                                                                                   // 37
  }                                                                                                          // 38
}                                                                                                            // 39
                                                                                                             // 40
fetchPicture = function (url) {                                                                              // 41
  try {                                                                                                      // 42
    var result = HTTP.get(url, {                                                                             // 43
      npmRequestOptions: { encoding: null },                                                                 // 44
      timeout: 5000                                                                                          // 45
    });                                                                                                      // 46
                                                                                                             // 47
    var metadata = {};                                                                                       // 48
                                                                                                             // 49
    metadata.mimeType = result.headers["content-type"];                                                      // 50
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                              // 51
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                             // 52
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                        // 53
    }                                                                                                        // 54
    var enc = result.headers["content-encoding"];                                                            // 55
    if (enc && enc !== "identity") {                                                                         // 56
      metadata.encoding = enc;                                                                               // 57
    }                                                                                                        // 58
                                                                                                             // 59
    return addStaticAsset(metadata, result.content);                                                         // 60
  } catch (err) {                                                                                            // 61
    console.error("failed to fetch user profile picture:", url, err.stack);                                  // 62
  }                                                                                                          // 63
}                                                                                                            // 64
                                                                                                             // 65
var ValidHandle = Match.Where(function (handle) {                                                            // 66
  check(handle, String);                                                                                     // 67
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                               // 68
});                                                                                                          // 69
                                                                                                             // 70
Accounts.onCreateUser(function (options, user) {                                                             // 71
  // The first non-dev user to sign in should be automatically upgraded to admin.                            // 72
  if (Meteor.users.find({"identities.service.dev": {$exists: 0}}).count() === 0 &&                           // 73
      !(options && options.service && options.service.dev)) {                                                // 74
    user.isAdmin = true;                                                                                     // 75
    user.signupKey = "admin";                                                                                // 76
  }                                                                                                          // 77
                                                                                                             // 78
  // Check profile.                                                                                          // 79
  if (options.profile) {                                                                                     // 80
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                     // 81
    check(options.profile, Match.ObjectIncluding({                                                           // 82
      name: Match.OneOf(null, Match.Optional(String)),                                                       // 83
      handle: Match.Optional(ValidHandle),                                                                   // 84
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                            // 85
    }));                                                                                                     // 86
  }                                                                                                          // 87
                                                                                                             // 88
  check(options.unverifiedEmail, Match.Optional(String));                                                    // 89
                                                                                                             // 90
  var identity = _.pick(options, "unverifiedEmail");                                                         // 91
  identity.profile = _.pick(options.profile || {}, "name", "handle", "pronouns", "picture");                 // 92
  identity.main = true;                                                                                      // 93
                                                                                                             // 94
  // Try downloading avatar.                                                                                 // 95
  var url = userPictureUrl(user);                                                                            // 96
  if (url) {                                                                                                 // 97
    var assetId = fetchPicture(url);                                                                         // 98
    if (assetId) {                                                                                           // 99
      identity.profile.picture = assetId;                                                                    // 100
    }                                                                                                        // 101
  }                                                                                                          // 102
                                                                                                             // 103
  var serviceUserId;                                                                                         // 104
  if (options.service && options.service.dev) {                                                              // 105
    check(options.service, {dev: {name: String}});                                                           // 106
    identity.service = options.service;                                                                      // 107
    serviceUserId = options.service.dev.name;                                                                // 108
  } else if ("expires" in user) {                                                                            // 109
    identity.service = {demo: {}};                                                                           // 110
    serviceUserId = user._id;                                                                                // 111
  } else if (options.service && options.service.email) {                                                     // 112
    check(options.service, {email:                                                                           // 113
                            {email: String,                                                                  // 114
                             tokens: [{digest: String, algorithm: String, createdAt: Date}]}});              // 115
    identity.service = options.service;                                                                      // 116
    identity.verifiedEmail = options.service.email.email;                                                    // 117
    serviceUserId = identity.verifiedEmail;                                                                  // 118
  } else if (user.services && "google" in user.services) {                                                   // 119
    identity.service = {google: {}};                                                                         // 120
    if (user.services.google.email && user.services.google.verified_email) {                                 // 121
      identity.verifiedEmail = user.services.google.email;                                                   // 122
    }                                                                                                        // 123
    serviceUserId = user.services.google.id;                                                                 // 124
  } else if (user.services && "github" in user.services) {                                                   // 125
    identity.service = {github: {}};                                                                         // 126
    if (user.services.github.email) {                                                                        // 127
      identity.unverifiedEmail = user.services.github.email;                                                 // 128
    }                                                                                                        // 129
    serviceUserId = user.services.github.id;                                                                 // 130
  }                                                                                                          // 131
  identity.id = Crypto.createHash("sha256")                                                                  // 132
      .update(Object.keys(identity.service)[0] + ":" + serviceUserId).digest("hex");                         // 133
                                                                                                             // 134
  user.identities = [identity];                                                                              // 135
  return user;                                                                                               // 136
});                                                                                                          // 137
                                                                                                             // 138
Accounts.validateLoginAttempt(function(info) {                                                               // 139
  if (info.allowed) {                                                                                        // 140
    if (info.type === "resume") {                                                                            // 141
      return true;                                                                                           // 142
    } else {                                                                                                 // 143
      var identities = info.user.identities;                                                                 // 144
      for (var ii = 0; ii < identities.length; ++ii) {                                                       // 145
        if (info.type in identities[ii].service) {                                                           // 146
          return !identities[ii].noLogin;                                                                    // 147
        }                                                                                                    // 148
      }                                                                                                      // 149
    }                                                                                                        // 150
  }                                                                                                          // 151
});                                                                                                          // 152
                                                                                                             // 153
Meteor.users._ensureIndex("identities.id", {unique: 1, sparse: 1});                                          // 154
                                                                                                             // 155
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/migrations.js                                                                       //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                          // 1
//                                                                                                           // 2
// Because migrations can experience partial failure and likely have                                         // 3
// side-effects, we should be careful to make sure all migrations are                                        // 4
// idempotent and safe to accidentally run multiple times.                                                   // 5
                                                                                                             // 6
var Future = Npm.require("fibers/future");                                                                   // 7
                                                                                                             // 8
var updateLoginStyleToRedirect = function() {                                                                // 9
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 10
  ["google", "github"].forEach(function(serviceName) {                                                       // 11
    var config = configurations.findOne({service: serviceName});                                             // 12
    if (config && config.loginStyle !== "redirect") {                                                        // 13
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                       // 14
    }                                                                                                        // 15
  });                                                                                                        // 16
};                                                                                                           // 17
                                                                                                             // 18
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                                 // 19
  // In the before time, Google and Github login were enabled by default.                                    // 20
  //                                                                                                         // 21
  // This actually didn't make much sense, required the first user to configure                              // 22
  // OAuth, and had some trust-the-first-user properties that weren't totally                                // 23
  // secure.                                                                                                 // 24
  //                                                                                                         // 25
  // Now, we have admin-token, so we wish to disable all logins by default                                   // 26
  // (since they need to be configured anyway) but since users may have never                                // 27
  // explicitly told Sandstorm that Google or Github login should be enabled,                                // 28
  // we can't just use the value in the Settings collection, since it might                                  // 29
  // never have been set.                                                                                    // 30
  //                                                                                                         // 31
  // Thus, if the service is configured but absent from Settings, we should                                  // 32
  // explicitly enable it in Settings, and then the rest of the logic can just                               // 33
  // depend on what value is in Settings and default to false without breaking                               // 34
  // user installations.                                                                                     // 35
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 36
  ["google", "github"].forEach(function(serviceName) {                                                       // 37
    var config = configurations.findOne({service: serviceName});                                             // 38
    var serviceConfig = Settings.findOne({_id: serviceName});                                                // 39
    if (config && !serviceConfig) {                                                                          // 40
      // Only explicitly enable the login service if:                                                        // 41
      // 1) the service is already configured                                                                // 42
      // 2) there is no sandstorm configuration already present (the user was                                // 43
      //    using the previous default behavior).                                                            // 44
      Settings.insert({_id: serviceName, value: true});                                                      // 45
    }                                                                                                        // 46
  });                                                                                                        // 47
};                                                                                                           // 48
                                                                                                             // 49
var denormalizeInviteInfo = function() {                                                                     // 50
  // When a user is invited via a signup token, the `signupKey` field of their user table entry              // 51
  // has always been populated to indicate the key they used. This points into the SignupKeys table          // 52
  // which has more information about the key, namely a freeform note entered by the admin when              // 53
  // they created the key. In the case that the email invite form was used, the note has the form            // 54
  // "E-mail invite to <address>".                                                                           // 55
  //                                                                                                         // 56
  // Later, we decided it was useful to indicate in the users table visible to the admin                     // 57
  // information about the invite terms. Namely, for email invites we want to show the address               // 58
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                // 59
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing              // 60
  // from the SignupKeys table.                                                                              // 61
                                                                                                             // 62
  Meteor.users.find().forEach(function (user) {                                                              // 63
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {              // 64
      var signupInfo = SignupKeys.findOne(user.signupKey);                                                   // 65
      if (signupInfo && signupInfo.note) {                                                                   // 66
        var newFields = { signupNote: signupInfo.note };                                                     // 67
                                                                                                             // 68
        var prefix = "E-mail invite to ";                                                                    // 69
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                     // 70
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                      // 71
        }                                                                                                    // 72
                                                                                                             // 73
        Meteor.users.update(user._id, {$set: newFields});                                                    // 74
      }                                                                                                      // 75
    }                                                                                                        // 76
  });                                                                                                        // 77
}                                                                                                            // 78
                                                                                                             // 79
function mergeRoleAssignmentsIntoApiTokens() {                                                               // 80
  RoleAssignments.find().forEach(function(roleAssignment) {                                                  // 81
    ApiTokens.insert({                                                                                       // 82
      grainId: roleAssignment.grainId,                                                                       // 83
      userId: roleAssignment.sharer,                                                                         // 84
      roleAssignment: roleAssignment.roleAssignment,                                                         // 85
      petname: roleAssignment.petname,                                                                       // 86
      created: roleAssignment.created,                                                                       // 87
      owner: {user: {userId: roleAssignment.recipient,                                                       // 88
                     title: roleAssignment.title}},                                                          // 89
    });                                                                                                      // 90
  });                                                                                                        // 91
}                                                                                                            // 92
                                                                                                             // 93
function fixOasisStorageUsageStats() {}                                                                      // 94
// This migration only pertained to Oasis and it was successfully applied there. Since it referred           // 95
// to some global variables that we later wanted to remove and/or rename, we've since replaced it            // 96
// with a no-op.                                                                                             // 97
                                                                                                             // 98
function fetchProfilePictures() {                                                                            // 99
  Meteor.users.find({}).forEach(function (user) {                                                            // 100
    var url = userPictureUrl(user);                                                                          // 101
    if (url) {                                                                                               // 102
      console.log("Fetching user picture:", url);                                                            // 103
      var assetId = fetchPicture(url);                                                                       // 104
      if (assetId) {                                                                                         // 105
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                                 // 106
      }                                                                                                      // 107
    }                                                                                                        // 108
  });                                                                                                        // 109
}                                                                                                            // 110
                                                                                                             // 111
function assignPlans() {                                                                                     // 112
  if (Meteor.settings.public.quotaEnabled && SandstormDb.paymentsMigrationHook) {                            // 113
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                                     // 114
  }                                                                                                          // 115
}                                                                                                            // 116
                                                                                                             // 117
function removeKeyrings() {                                                                                  // 118
  // These blobs full of public keys were not intended to find their way into mongo and while                // 119
  // harmless they slow things down because they're huge. Remove them.                                       // 120
  Packages.update({"manifest.metadata.pgpKeyring": {$exists: true}},                                         // 121
      {$unset: {"manifest.metadata.pgpKeyring": ""}},                                                        // 122
      {multi: true});                                                                                        // 123
}                                                                                                            // 124
                                                                                                             // 125
function useLocalizedTextInUserActions() {                                                                   // 126
  function toLocalizedText(newObj, oldObj, field) {                                                          // 127
    if (field in oldObj) {                                                                                   // 128
      if (typeof oldObj[field] === "string") {                                                               // 129
        newObj[field] = {defaultText: oldObj[field]};                                                        // 130
      } else {                                                                                               // 131
        newObj[field] = oldObj[field];                                                                       // 132
      }                                                                                                      // 133
    }                                                                                                        // 134
  }                                                                                                          // 135
  UserActions.find({}).forEach(function (userAction) {                                                       // 136
    var fields = {};                                                                                         // 137
    toLocalizedText(fields, userAction, "appTitle");                                                         // 138
    toLocalizedText(fields, userAction, "title");                                                            // 139
    toLocalizedText(fields, userAction, "nounPhrase");                                                       // 140
    UserActions.update(userAction._id, {$set: fields});                                                      // 141
  });                                                                                                        // 142
}                                                                                                            // 143
                                                                                                             // 144
function verifyAllPgpSignatures() {                                                                          // 145
  Packages.find({}).forEach(function (pkg) {                                                                 // 146
    try {                                                                                                    // 147
      console.log("checking PGP signature for package:", pkg._id);                                           // 148
      var info = waitPromise(globalBackend.cap().tryGetPackage(pkg._id));                                    // 149
      if (info.authorPgpKeyFingerprint) {                                                                    // 150
        console.log("  " + info.authorPgpKeyFingerprint);                                                    // 151
        Packages.update(pkg._id,                                                                             // 152
            {$set: {authorPgpKeyFingerprint: info.authorPgpKeyFingerprint}});                                // 153
      } else {                                                                                               // 154
        console.log("  no signature");                                                                       // 155
      }                                                                                                      // 156
    } catch (err) {                                                                                          // 157
      console.error(err.stack);                                                                              // 158
    }                                                                                                        // 159
  });                                                                                                        // 160
}                                                                                                            // 161
                                                                                                             // 162
function splitUserIdsIntoAccountIdsAndIdentityIds() {                                                        // 163
  var Crypto = Npm.require("crypto");                                                                        // 164
  Meteor.users.find().forEach(function (user) {                                                              // 165
    var identity = {};                                                                                       // 166
    var serviceUserId;                                                                                       // 167
    if ("devName" in user) {                                                                                 // 168
      identity.service = "dev";                                                                              // 169
      serviceUserId = user.devName;                                                                          // 170
    } else if ("expires" in user) {                                                                          // 171
      identity.service = "demo";                                                                             // 172
      serviceUserId = user._id;                                                                              // 173
    } else if (user.services && "google" in user.services) {                                                 // 174
      identity.service = "google";                                                                           // 175
      if (user.services.google.email && user.services.google.verified_email) {                               // 176
        identity.verifiedEmail = user.services.google.email;                                                 // 177
      }                                                                                                      // 178
      serviceUserId = user.services.google.id;                                                               // 179
    } else if (user.services && "github" in user.services) {                                                 // 180
      identity.service = "github";                                                                           // 181
      identity.unverifiedEmail = user.services.github.email;                                                 // 182
      serviceUserId = user.services.github.id;                                                               // 183
    } else if (user.services && "emailToken" in user.services) {                                             // 184
      identity.service = "emailToken";                                                                       // 185
      identity.verifiedEmail = user.services.emailToken.email;                                               // 186
      serviceUserId = user.services.emailToken.email;                                                        // 187
    }                                                                                                        // 188
                                                                                                             // 189
    identity.id = Crypto.createHash("sha256")                                                                // 190
        .update(identity.service + ":" + serviceUserId).digest("hex");                                       // 191
                                                                                                             // 192
    if (user.profile) {                                                                                      // 193
      if (user.profile.name) {                                                                               // 194
        identity.name = user.profile.name;                                                                   // 195
      }                                                                                                      // 196
      if (user.profile.handle) {                                                                             // 197
        identity.handle = user.profile.handle;                                                               // 198
      }                                                                                                      // 199
      if (user.profile.picture) {                                                                            // 200
        identity.picture = user.profile.picture;                                                             // 201
      }                                                                                                      // 202
      if (user.profile.pronoun) {                                                                            // 203
        identity.pronoun = user.profile.pronoun;                                                             // 204
      }                                                                                                      // 205
      if (user.profile.email) {                                                                              // 206
        identity.unverifiedEmail = user.profile.email;                                                       // 207
      }                                                                                                      // 208
    }                                                                                                        // 209
    identity.main = true;                                                                                    // 210
                                                                                                             // 211
    Meteor.users.update(user._id, {$set: {identities: [identity]}});                                         // 212
                                                                                                             // 213
    Grains.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                     // 214
    Sessions.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                   // 215
    ApiTokens.update({userId: user._id},                                                                     // 216
                     {$set: {identityId: identity.id}},                                                      // 217
                     {multi: true});                                                                         // 218
    ApiTokens.update({"owner.user.userId": user._id},                                                        // 219
                     {$set: {"owner.user.identityId": identity.id}},                                         // 220
                     {multi: true});                                                                         // 221
    ApiTokens.update({"owner.grain.introducerUser": user._id},                                               // 222
                     {$set: {"owner.grain.introducerIdentity": identity.id}},                                // 223
                     {multi: true});                                                                         // 224
                                                                                                             // 225
    while (ApiTokens.update({"requirements.permissionsHeld.userId": user._id},                               // 226
                            {$set: {"requirements.$.permissionsHeld.identityId": identity.id},               // 227
                             $unset: {"requirements.$.permissionsHeld.userId": 1}},                          // 228
                            {multi: true}) > 0);                                                             // 229
    // The `$` operatorer modifies the first element in the array that matches the query. Since              // 230
    // there may be many matches, we need to repeat until no documents are modified.                         // 231
                                                                                                             // 232
  });                                                                                                        // 233
                                                                                                             // 234
  ApiTokens.remove({userInfo: {$exists: true}});                                                             // 235
  // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                    // 236
  // that this field could show up in the database was in this deprecated, no-longer-functional              // 237
  // form of API token.                                                                                      // 238
}                                                                                                            // 239
                                                                                                             // 240
function appUpdateSettings() {                                                                               // 241
  Settings.insert({_id: "appMarketUrl", value: "https://apps.sandstorm.io"});                                // 242
  Settings.insert({_id: "appIndexUrl", value: "https://app-index.sandstorm.io"});                            // 243
  Settings.insert({_id: "appUpdatesEnabled", value: true});                                                  // 244
}                                                                                                            // 245
                                                                                                             // 246
function moveDevAndEmailLoginDataIntoIdentities() {                                                          // 247
  Meteor.users.find().forEach(function (user) {                                                              // 248
    if (user.identities.length != 1) {                                                                       // 249
      throw new Error("User does not have exactly one identity: ", user);                                    // 250
    }                                                                                                        // 251
    var identity = user.identities[0];                                                                       // 252
    if (Match.test(identity.service, Object)) { return; } // Already migrated.                               // 253
                                                                                                             // 254
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");         // 255
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                          // 256
                                                                                                             // 257
    var serviceObject = {};                                                                                  // 258
    var fieldsToUnset = {}                                                                                   // 259
                                                                                                             // 260
    if (identity.service === "dev") {                                                                        // 261
      serviceObject.name = user.devName;                                                                     // 262
      fieldsToUnset.devName = 1;                                                                             // 263
    } else if (identity.service === "emailToken") {                                                          // 264
      serviceObject.tokens = user.services.emailToken.tokens;                                                // 265
      serviceObject.email = user.services.emailToken.email;                                                  // 266
      fieldsToUnset["services.emailToken"] = 1;                                                              // 267
    }                                                                                                        // 268
                                                                                                             // 269
    newIdentity.service = {};                                                                                // 270
    newIdentity.service[identity.service] = serviceObject;                                                   // 271
                                                                                                             // 272
    var modifier = {$set: {identities: [newIdentity]}};                                                      // 273
    if (Object.keys(fieldsToUnset).length > 0) {                                                             // 274
      modifier["$unset"] = fieldsToUnset;                                                                    // 275
    }                                                                                                        // 276
                                                                                                             // 277
    Meteor.users.update({_id: user._id}, modifier);                                                          // 278
  });                                                                                                        // 279
}                                                                                                            // 280
                                                                                                             // 281
function repairEmailIdentityIds() {                                                                          // 282
  var Crypto = Npm.require("crypto");                                                                        // 283
  Meteor.users.find({"identities.service.emailToken": {$exists: 1}}).forEach(function (user) {               // 284
    if (user.identities.length != 1) {                                                                       // 285
      throw new Error("User does not have exactly one identity: ", user);                                    // 286
    }                                                                                                        // 287
    var identity = user.identities[0];                                                                       // 288
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail",                 // 289
                             "profile");                                                                     // 290
    newIdentity.service = {email: identity.service.emailToken};                                              // 291
    newIdentity.id = Crypto.createHash("sha256")                                                             // 292
      .update("email:" + identity.service.emailToken.email).digest("hex");                                   // 293
                                                                                                             // 294
    Grains.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});           // 295
    Sessions.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});         // 296
    ApiTokens.update({identityId: identity.id},                                                              // 297
                     {$set: {identityId: newIdentity.id}},                                                   // 298
                     {multi: true});                                                                         // 299
    ApiTokens.update({"owner.user.identityId": identity.id},                                                 // 300
                     {$set: {"owner.user.identityId": newIdentity.id}},                                      // 301
                     {multi: true});                                                                         // 302
    ApiTokens.update({"owner.grain.introducerIdentity": identity.id},                                        // 303
                     {$set: {"owner.grain.introducerIdentity": newIdentity.id}},                             // 304
                     {multi: true});                                                                         // 305
                                                                                                             // 306
    while (ApiTokens.update({"requirements.permissionsHeld.identityId": identity.id},                        // 307
                            {$set: {"requirements.$.permissionsHeld.identityId": newIdentity.id}},           // 308
                            {multi: true}) > 0);                                                             // 309
                                                                                                             // 310
    Meteor.users.update({_id: user._id}, {$set: {identities: [newIdentity]}});                               // 311
  });                                                                                                        // 312
}                                                                                                            // 313
                                                                                                             // 314
// This must come after all the functions named within are defined.                                          // 315
// Only append to this list!  Do not modify or remove list entries;                                          // 316
// doing so is likely change the meaning and semantics of user databases.                                    // 317
var MIGRATIONS = [                                                                                           // 318
  updateLoginStyleToRedirect,                                                                                // 319
  enableLegacyOAuthProvidersIfNotInSettings,                                                                 // 320
  denormalizeInviteInfo,                                                                                     // 321
  mergeRoleAssignmentsIntoApiTokens,                                                                         // 322
  fixOasisStorageUsageStats,                                                                                 // 323
  fetchProfilePictures,                                                                                      // 324
  assignPlans,                                                                                               // 325
  removeKeyrings,                                                                                            // 326
  useLocalizedTextInUserActions,                                                                             // 327
  verifyAllPgpSignatures,                                                                                    // 328
  splitUserIdsIntoAccountIdsAndIdentityIds,                                                                  // 329
  appUpdateSettings,                                                                                         // 330
  moveDevAndEmailLoginDataIntoIdentities,                                                                    // 331
  repairEmailIdentityIds,                                                                                    // 332
];                                                                                                           // 333
                                                                                                             // 334
function migrateToLatest() {                                                                                 // 335
  if (Meteor.settings.replicaNumber) {                                                                       // 336
    // This is a replica. Wait for the first replica to perform migrations.                                  // 337
                                                                                                             // 338
    console.log("Waiting for migrations on replica zero...");                                                // 339
                                                                                                             // 340
    var done = new Future();                                                                                 // 341
    var change = function (doc) {                                                                            // 342
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                   // 343
      if (doc.value >= MIGRATIONS.length) done.return();                                                     // 344
    }                                                                                                        // 345
    var observer = Migrations.find({_id: "migrations_applied"}).observe({                                    // 346
      added: change,                                                                                         // 347
      changed: change                                                                                        // 348
    });                                                                                                      // 349
                                                                                                             // 350
    done.wait();                                                                                             // 351
    observer.stop();                                                                                         // 352
    console.log("Migrations have completed on replica zero.");                                               // 353
                                                                                                             // 354
  } else {                                                                                                   // 355
    var applied = Migrations.findOne({_id: "migrations_applied"});                                           // 356
    var start;                                                                                               // 357
    if (!applied) {                                                                                          // 358
      // Migrations table is not yet seeded with a value.  This means it has                                 // 359
      // applied 0 migrations.  Persist this.                                                                // 360
      Migrations.insert({_id: "migrations_applied", value: 0});                                              // 361
      start = 0;                                                                                             // 362
    } else {                                                                                                 // 363
      start = applied.value;                                                                                 // 364
    }                                                                                                        // 365
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                           // 366
                                                                                                             // 367
    for (var i = start ; i < MIGRATIONS.length ; i++) {                                                      // 368
      // Apply migration i, then record that migration i was successfully run.                               // 369
      console.log("Applying migration " + (i+1));                                                            // 370
      MIGRATIONS[i]();                                                                                       // 371
      Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                                  // 372
      console.log("Applied migration " + (i+1));                                                             // 373
    }                                                                                                        // 374
  }                                                                                                          // 375
}                                                                                                            // 376
                                                                                                             // 377
SandstormDb.prototype.migrateToLatest = migrateToLatest;                                                     // 378
                                                                                                             // 379
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
